"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Population import Population
from ANNarchy.core.PopulationView import PopulationView
from ANNarchy.core.Projection import Projection
from ANNarchy.core.Dendrite import Dendrite
from ANNarchy.core import Global

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages

import numpy as np
import re
import sys
from copy import copy, deepcopy
from typing import Any
import h5py

# objects/functions that should be available by "from ANNarchy import *"
__all__ = ["Monitor"]

class Monitor :
    """
    Object allowing to record variables from `Population`, `PopulationView`, `Dendrite` or `Projection` instances.

    This object should not be created directly, but returned by `Network.monitor()`:

    ```python
    m = net.monitor(pop, ['g_exc', 'v', 'spike'], period=10.0, start=False)
    ```

    Monitors are started by default after `compile()`. You can control their recording behavior with the `start()`, `stop()`, `pause()` and `resume()` methods.

    ```python
    m.start() # Start recording
    net.simulate(T)
    m.pause() # Pause recording
    net.simulate(T)
    m.resume() # Resume recording
    net.simulate(T)

    data = m.get() # Get the data
    ```

    For spiking networks recording `'spike'`, some utilities allow to easily compute raster plots /other statistics or mean firing rates over time/neuron axes:

    ```python
    spikes = m.get('spike')
    
    t, n = m.raster_plot(spikes)
    histo = m.histogram()
    isi = m.inter_spike_interval(spikes)
    cov = m.coefficient_of_variation(spikes)
    fr = m.mean_fr(spikes)
    r = m.smoothed_rate(spikes, smooth=100.)
    r_mean = m.population_rate(spikes, smooth=100.)
    ```
    """

    def __init__(self,
                 obj: Any,
                 variables:list=[],
                 period:float=None,
                 period_offset:float=None,
                 start:bool=True,
                 name:str=None,
                 net_id:int=0):


        # Object to record (Population, PopulationView, Dendrite)
        self.object = obj
        self.cyInstance = None
        self.net_id = net_id

        # Check type of the object
        if not isinstance(self.object, (Population, Projection, PopulationView, Dendrite)):
            Messages._error('Monitor: the object must be a Population, PopulationView, Dendrite or Projection object')

        # dt is saved in the network
        self.dt = NetworkManager().get_network(net_id=net_id).dt

        # Get a name
        self.name = name
        if self.name is None:
            if isinstance(self.object, (Population, Projection)):
                self.name = 'Monitor_'+obj.name
            elif isinstance(self.object, PopulationView):
                self.name = 'Monitor_'+obj.population.name
            elif isinstance(self.object, Dendrite):
                self.name = 'Monitor_'+obj.proj.name

        # Variables to record
        if not isinstance(variables, list):
            self._variables = [variables]
        else:
            self._variables = variables

        # Sanity check: we want only record variables
        for var in self._variables:
            if var == "w" and var in self.object.variables:
                continue

            if var in self.object.parameters:
                Messages._error('Parameters are not recordable')

            if not var in self.object.variables and var not in ['spike', 'axon_spike'] and not var.startswith('sum('):
                Messages._error('Monitor: the object does not have an attribute named', var)

        # Period
        if not period:
            self._period = self.dt
        else:
            self._period = float(period)

        # Period Offset
        if not period_offset:
            self._period_offset = 0
        else:
            # Check validity
            if period_offset >= period:
                Messages._error("Monitor(): value of period_offset must be smaller than period.")
            else:
                self._period_offset = period_offset

        # Warn users when recording projections all the time
        if isinstance(self.object, Projection) and self._period == self.dt:
            Messages._warning('Monitor(): it is a bad idea to record synaptic variables of a projection at each time step!')

        # Start
        self._start = start
        self._recorded_variables = {}
        self._last_recorded_variables = {}

        # Add the monitor to the global variable
        self.id = NetworkManager().get_network(net_id=net_id)._add_monitor(self)

        if NetworkManager().get_network(net_id=net_id).compiled: # Already compiled
            self._init_monitoring()

    def _copy(self, net_id=None):
        "Returns a copy of the monitor when creating networks. Internal use only."

        return Monitor(
                obj=self.object, 
                variables=self._variables, 
                period=self._period, 
                period_offset=self._period_offset, 
                start=self._start, 
                name=self.name,
                net_id=self.net_id if net_id is None else net_id,
            )

    # Extend the period attribute
    @property
    def period(self) -> float:
        "Period of recording in milliseconds."
        if not self.cyInstance:
            return self._period
        else:
            return self.cyInstance.period * ConfigManager().get('dt', net_id=self.net_id)
    @period.setter
    def period(self, val):
        if not self.cyInstance:
            self._period = val
        else:
            self.cyInstance.period = int(val/ConfigManager().get('dt', self.net_id))

    # Extend the period_offset attribute
    @property
    def period_offset(self) -> float:
        "Offset of recording within a period, in milliseconds."
        if not self.cyInstance:
            return self._period
        else:
            return self.cyInstance.period_offset * ConfigManager().get('dt', self.net_id)

    @period_offset.setter
    def period_offset(self, val):
        if not self.cyInstance:
            self._period = val
        else:
            self.cyInstance.period_offset = int(val/ConfigManager().get('dt', self.net_id))

    # Extend the variables attribute
    @property
    def variables(self) -> list:
        "Current list of recorded variables."
        return copy(self._variables)

    @variables.setter
    def variables(self, val):
        Messages._error("Modifying of a Monitors variable list is not allowed")



    def get(self, 
            variables:str | list[str]=None, 
            keep:bool=False, 
            reshape:bool=False, 
            force_dict:bool=False
        ) -> dict:
        """
        Returns the recorded variables and empties the buffer.
         
        The recorded data is returned as a Numpy array (first dimension is time, second is neuron index).

        If a single variable name is provided, the recorded values for this variable are directly returned as an array.
        If a list is provided or the argument left empty, a dictionary with all recorded variables is returned.

        The `spike` variable of a population will be returned as a dictionary of lists, where the key is the neuron index, and the list contains the spike times (in **steps**; multiply by `net.dt` to get spike times in milliseconds) for each recorded neurons.

        :param variables: (list of) variables. By default, a dictionary with all variables is returned.
        :param keep: defines if the content in memory for each variable should be kept (default: False).
        :param reshape: transforms the second axis of the array to match the population's geometry (default: False).
        """
        if variables:
            if not isinstance(variables, list):
                variables = [variables]
        else:
            variables = self.variables
            force_dict = True

        data = {}
        for var in variables:
            name = var
            # Sums of inputs for rate-coded populations
            if var.startswith('sum('):
                target = re.findall(r"\(([\w]+)\)", var)[0]
                name = '_sum_' + target

            # Retrieve the data
            data[var] = self._return_variable(name, keep, reshape)

            # Update stopping time
            self._update_stopping_time(var, keep)

        if not force_dict and len(variables)==1:
            return data[variables[0]]
        else:
            return data



    def _size_in_bytes(self) -> int:
        """
        Returns the size of allocated memory on C++ side. This is only valid if compile() was invoked.
        """
        if hasattr(self.cyInstance, 'size_in_bytes'):
            return self.cyInstance.size_in_bytes()

    def _clear(self):
        """
        Deallocates the container within the C++ instance. The population object is not usable anymore after calling this function.

        Warning: should be only called by the net deconstructor (in the context of parallel_run).
        """
        if hasattr(self.cyInstance, 'clear'):
            self.cyInstance.clear()

    def _add_variable(self, var):
        """
        Adds a variable to the list of recorded attributes.
        """
        if not var in self._variables:
            self._variables.append(var)

        self._recorded_variables[var] = {
            'start': [Global.get_current_step(self.net_id)],
            'stop': [None],
        }

        self._last_recorded_variables[var] = {
            'start': [Global.get_current_step(self.net_id)],
            'stop': [None],
        }


    def _init_monitoring(self):
        "To be called after compile() as it accesses cython objects"
        # Start recording dependent on the recorded object
        from ANNarchy.extensions.bold import BoldMonitor
        if isinstance(self, BoldMonitor):
            self._start_bold_monitor() # pylint: disable=no-member
        elif isinstance(self.object, (Population, PopulationView)):
            self._start_population()
        elif isinstance(self.object, (Dendrite, Projection)):
            self._start_dendrite()

    def _start_population(self):
        "Creates the C++ object and starts the recording for a population."

        if isinstance(self.object, PopulationView):
            self.ranks = list(self.object.ranks)
        else:
            self.ranks = [-1]

        # Create the wrapper
        period = int(self._period/ConfigManager().get('dt', self.net_id))
        period_offset = int(self._period_offset/ConfigManager().get('dt', self.net_id))
        offset = Global.get_current_step(self.net_id) % period

        # Create the instance
        self.cyInstance = getattr(NetworkManager().get_network(net_id=self.net_id).instance, 'PopRecorder'+str(self.object.id)+'_wrapper')(self.ranks, period, period_offset, offset)

        # Add variables
        for var in self._variables:
            self._add_variable(var)

        # Start recordings if enabled
        if self._start:
            self.start()

    def _start_dendrite(self):
        "Creates the C++ object and starts the recording for a dendrite."

        if isinstance(self.object, Dendrite):
            self.ranks = self.object.post_rank
            self.idx = [self.object.idx]
            proj_id = self.object.proj.id
        else: # Projection
            self.ranks = [-1]
            self.idx = self.object.post_ranks
            proj_id = self.object.id

        # Compute the period and offset
        period = int(self._period/ConfigManager().get('dt', self.net_id))
        period_offset = int(self._period_offset / ConfigManager().get('dt', self.net_id))
        offset = Global.get_current_step(self.net_id) % period

        # Create the wrapper
        self.cyInstance = getattr(NetworkManager().get_network(net_id=self.net_id).instance, 'ProjRecorder'+str(proj_id)+'_wrapper')(self.idx, period, period_offset, offset)

        # Add the variables
        for var in self._variables:
            self._add_variable(var)

        # Start recordings if enabled
        if self._start:
            self.start()

    def start(self, variables:list=None, period:float=None) -> None:
        """
        Starts recording the variable.

        It is called automatically after ``Network.compile()`` if the flag ``start=False`` was not passed to the constructor.

        :param variables: single variable name or list of variable names to start recording (default: the ``variables`` argument passed to the constructor).
        :param period: delay in ms between two recording (default: dt). Not valid for the ``spike`` variable of a Population(View).
        """
        if variables:
            if not isinstance(variables, list):
                self._add_variable(variables)
                variables = [variables]
            else:
                for var in variables:
                    self._add_variable(var)
        else:
            variables = self.variables

        if period:
            self._period = period
            self.cyInstance.period = int(self._period/ConfigManager().get('dt', self.net_id))
            self.cyInstance.offset = Global.get_current_step(self.net_id)

        for var in variables:
            name = var
            # Sums of inputs for rate-coded populations
            if var.startswith('sum('):
                target = re.findall(r"\(([\w]+)\)", var)[0]
                name = '_sum_' + target
            try:
                setattr(self.cyInstance, 'record_'+name, True)
            except:
                obj_desc = ''
                if isinstance(self.object, (Population, PopulationView)):
                    obj_desc = 'population ' + self.object.name
                elif isinstance(self.object, Projection):
                    obj_desc = 'projection between '+self.object.pre.name+' and '+self.object.post.name
                else:
                    obj_desc = 'dendrite between '+self.object.proj.pre.name+' and '+self.object.proj.post.name
                    if var in self.object.proj.parameters:
                        Messages._print('\t', var, 'is a parameter, its value is constant')

                Messages._warning('Monitor: ' + var + ' can not be recorded ('+obj_desc+')')


    def pause(self) -> None:
        "Pauses the recording."
        # Start recording the variables
        for var in self.variables:
            name = var
            # Sums of inputs for rate-coded populations
            if var.startswith('sum('):
                target = re.findall(r"\(([\w]+)\)", var)[0]
                name = '_sum_' + target
            try:
                setattr(self.cyInstance, 'record_'+name, False)
            except:
                obj_desc = ''
                if isinstance(self.object, (Population, PopulationView)):
                    obj_desc = 'population ' + self.object.name
                elif isinstance(self.object, Projection):
                    obj_desc = 'projection between ' + self.object.pre.name+' and '+self.object.post.name
                else:
                    obj_desc = 'dendrite between '+self.object.proj.pre.name+' and '+self.object.proj.post.name
                Messages._warning('Monitor:' + var + ' can not be recorded ('+obj_desc+')')

            self._recorded_variables[var]['stop'][-1] = Global.get_current_step(self.net_id)


    def resume(self) -> None:
        "Resumes the recording."
        # Start recording the variables
        for var in self.variables:
            name = var
            # Sums of inputs for rate-coded populations
            if var.startswith('sum('):
                target = re.findall(r"\(([\w]+)\)", var)[0]
                name = '_sum_' + target
            try:
                setattr(self.cyInstance, 'record_'+name, True)
            except:
                obj_desc = ''
                if isinstance(self.object, (Population, PopulationView)):
                    obj_desc = 'population '+self.object.name
                elif isinstance(self.object, Projection):
                    obj_desc = 'projection between '+self.object.pre.name+' and '+self.object.post.name
                else:
                    obj_desc = 'dendrite between '+self.object.proj.pre.name+' and '+self.object.proj.post.name
                Messages._warning('Monitor:' + var + ' can not be recorded ('+obj_desc+')')

            self._recorded_variables[var]['start'].append(Global.get_current_step(self.net_id))
            self._recorded_variables[var]['stop'].append(None)

    def stop(self) -> None:
        """
        Stops the recording.

        Warning: This will delete the content of the C++ object and all data not previously retrieved is lost.
        """
        try:
            self._variables = []
            self._recorded_variables = {}
            self.cyInstance.clear()
            self.cyInstance = None

        except:
            obj_desc = ''
            if isinstance(self.object, (Population, PopulationView)):
                obj_desc = 'population '+self.object.name
            elif isinstance(self.object, Projection):
                obj_desc = 'projection between '+self.object.pre.name+' and '+self.object.post.name
            else:
                obj_desc = 'dendrite between '+self.object.proj.pre.name+' and '+self.object.proj.post.name
            Messages._warning('Monitor:' + obj_desc + 'cannot be stopped')

    def reset(self) -> None:
        """
        Reset the monitor to its initial state.
        """
        for var in self._variables:
            # Flush the data
            data = self.get(var)
            del data
            # Reinitializes the timings
            self._add_variable(var)


    def _return_variable(self, name, keep, reshape):
        """ Returns the value of a variable with the given name. """
        
        if isinstance(self.object, (Population, PopulationView)):
            if not reshape:
                return self._get_population(self.object, name, keep)
            return np.reshape(self._get_population(self.object, name, keep), (-1,) + self.object.geometry)
        
        if isinstance(self.object, Projection):
            return self._get_dendrite(self.object, name, keep)
        
        if isinstance(self.object, Dendrite):
            # Dendrites have one empty dimension
            return self._get_dendrite(self.object, name, keep).squeeze()
        
        return None

    def _update_stopping_time(self, var, keep):
        
        self._recorded_variables[var]['stop'][-1] = Global.get_current_step(self.net_id)
        self._last_recorded_variables[var]['start'] = self._recorded_variables[var]['start']
        self._last_recorded_variables[var]['stop'] = self._recorded_variables[var]['stop']
        
        if not keep:
            self._recorded_variables[var]['start'] = [Global.get_current_step(self.net_id)]
            self._recorded_variables[var]['stop'] = [None]

    
    def __getitem__(self, key):
        # Implement the logic to retrieve the item by key
        return self.get(key)

    def save(self, filename:str, variables:str | list[str]=None,
             keep:bool=False, reshape:bool=False, force_dict:bool=False) -> None:
        """
        Saves the recorded variables as a Numpy array (first dimension is time, second is neuron index).

        If a single variable name is provided, the recorded values for this variable are directly saved.
        If a list is provided or the argument left empty, a dictionary with all recorded variables is saved.

        The `spike` variable of a population will be returned as a dictionary of lists containing the spike times (in **steps**; multiply by `net.dt` to get spike times in milliseconds) for each recorded neurons.

        :param filename: name of the save file.
        :param variables: (list of) variables. By default, a dictionary with all variables is returned.
        :param keep: defines if the content in memory for each variable should be kept (default: False).
        :param reshape: transforms the second axis of the array to match the population's geometry (default: False).
        """
        if variables:
            if not isinstance(variables, list):
                variables = [variables]
        else:
            variables = self.variables
            force_dict = True

        ## Save single variables as numpy array
        if filename.endswith(".npy"):
            if len(variables) == 1:
                Messages._error('Monitor.save: Saving with numpy only possible for single variables.')
            name = variables[0]
            # Sums of inputs for rate-coded populations
            if name.startswith('sum('):
                target = re.findall(r"\(([\w]+)\)", name)[0]
                name = '_sum_' + target

            # Retrieve the data
            np.save(filename, self._return_variable(name, keep, reshape))

            # Update stopping time
            self._update_stopping_time(variables[0], keep)
        elif filename.endswith(".hdf5"):
            ## Save as multiple variables as h5py File
            with h5py.File(filename, 'w') as data:
                for var in variables:
                    name = var
                    # Sums of inputs for rate-coded populations
                    if var.startswith('sum('):
                        target = re.findall(r"\(([\w]+)\)", var)[0]
                        name = '_sum_' + target

                    # Retrieve the data
                    data["/" + var] = self._return_variable(name, keep, reshape)

                    # Update stopping time
                    self._update_stopping_time(var, keep)
        else:
            Messages._error('Monitor.save: File type not recognized (Must be .hdf5 or .npy).')

    def _get_population(self, pop, name, keep):
        try:
            data = getattr(self.cyInstance, name)
            if not keep:
                getattr(self.cyInstance, 'clear_' + name)()
        except:
            data = []

        if name not in ['spike', 'axon_spike']:
            return np.array(data)
        else:
            return data

    def _get_dendrite(self, proj, name, keep):
        try:
            data = getattr(self.cyInstance, name)
            if not keep:
                getattr(self.cyInstance, 'clear_' + name)()
        except:
            data = []
        
        return np.array(data, dtype=object)

    def times(self, variables:list[str]=None) -> dict:
        """
        Returns the start and stop times (in ms) of the recorded variables as a dictionary.

        It should only be called after a call to ``get()``, so that it describes when the variables have been recorded.

        :param variables: (list of) variables. By default, the times for all variables is returned.
        """
        t = {}
        if variables:
            if not isinstance(variables, list):
                variables = [variables]
        else:
            variables = self._variables

        for var in variables:
            # check for spelling mistakes
            if not var in self._variables:
                Messages._warning("Variable '"+str(var)+"' is not monitored.")
                continue

            t[var] = deepcopy(self._last_recorded_variables[var])

        return t

    ###############################
    ### Spike visualisation stuff
    ###############################
    def raster_plot(self, spikes:dict=None) -> tuple:
        """
        Returns two numpy arrays representing for each recorded spike 1) the spike times and 2) the ranks of the neurons.

        Example:

        ```python
        m = net.monitor(pop, 'spike')
        net.simulate(1000.0)
        
        t, n = m.raster_plot()
        plt.plot(t, n, '.')
        ```

        or:

        ```python
        m = net.monitor(pop, 'spike')
        net.simulate(1000.0)

        spikes = m.get('spike')
        t, n = m.raster_plot(spikes)
        plt.plot(t, n, '.')
        ```

        :param spikes: the dictionary of spikes returned by ``get('spike')``. If left empty, ``get('spike')`` will be called. Beware: this erases the data from memory.
        """
        times = []; ranks=[]
        if not 'spike' in self._variables:
            Messages._error('Monitor: spike was not recorded')

        # Get data
        if spikes is None:
            data = self.get('spike')
        else:
            if 'spike' in spikes.keys():
                data = spikes['spike']
            elif 'axon_spike' in spikes.keys():
                data = spikes['axon_spike']
            else:
                data = spikes

        # Compute raster
        for n in data.keys():
            for t in data[n]:
                times.append(t)
                ranks.append(n)

        return self.dt * np.array(times), np.array(ranks)

    def histogram(self, spikes=None, bins=None, per_neuron=False, recording_window=None):
        """
        Returns a histogram for the recorded spikes in the population.

        ```python
        m = net.monitor(pop, 'spike')
        net.simulate(1000.0)

        spikes = m.get('spike')
        histo = m.histogram(spikes)
        plt.plot(histo)
        ```

        :param spikes: the dictionary of spikes returned by ``get('spike')``. If left empty, ``get('spike')`` will be called. Beware: this erases the data from memory.
        :param bins: the bin size in ms (default: dt).
        """
        if not 'spike' in self._variables:
            Messages._error('Monitor: spike was not recorded')

        # Get data
        if not spikes:
            data = self.get('spike')
        else:
            if 'spike' in spikes.keys():
                data = spikes['spike']
            else:
                data = spikes

        return histogram(data, bins=bins, per_neuron=per_neuron, recording_window=recording_window, dt=self.dt)

    def inter_spike_interval(self, spikes:dict=None, ranks:list[int]=None, per_neuron:bool=False) -> list:
        """
        Computes the inter-spike intervals (ISI) for the recorded spikes in the population.

        ```python
        m = net.monitor(pop, 'spike')
        net.simulate(1000.0)

        spikes = m.get('spike')
        isi = m.inter_spike_interval(spikes)
        plt.hist(isi)
        ```

        :param spikes: the dictionary of spikes returned by ``get('spike')``. If left empty, ``get('spike')`` will be called. Beware: this erases the data from memory.
        :param ranks: a list of neurons that should be evaluated. By default `None`, all neurons are evaluated.
        :param per_neuron: if set to True, the computed inter-spike intervals are stored per neuron (analog to spikes), otherwise all values are stored in one huge vector (default: False).
        """
        # Get data
        if not spikes:
            data = self.get('spike')
        else:
            if 'spike' in spikes.keys():
                data = spikes['spike']
            else:
                data = spikes

        return inter_spike_interval(data, ranks=ranks, per_neuron=per_neuron, dt=self.dt)

    def coefficient_of_variation(self, spikes:dict=None, ranks:list[int]=None) -> list:
        """
        Computes the coefficient of variation for the recorded spikes in the population.
   
        ```python
        m = net.monitor(pop, 'spike')
        net.simulate(1000.0)

        spikes = m.get('spike')
        cov = m.coefficient_of_variation(spikes)
        plt.hist(isi)
        ```

        :param spikes: the dictionary of spikes returned by ``get('spike')``. If left empty, ``get('spike')`` will be called. Beware: this erases the data from memory.
        :ranks: a list of neurons that should be evaluated. By default (None), all neurons are evaluated.
        """
        # Get data
        if not spikes:
            data = self.get('spike')
        else:
            if 'spike' in spikes.keys():
                data = spikes['spike']
            else:
                data = spikes

        return coefficient_of_variation(data, ranks=ranks, dt=self.dt)

    def mean_fr(self, spikes:dict=None) -> float:
        """
        Computes the mean firing rate in the population during the recordings.

        ```python
        m = net.monitor(pop, 'spike')
        net.simulate(1000.0)

        spikes = m.get('spike')
        fr = m.mean_fr(spikes)
        ```

        :param spikes: the dictionary of spikes returned by ``get('spike')``. If left empty, ``get('spike')`` will be called. Beware: this erases the data from memory.

        """
        if not 'spike' in self._variables:
            Messages._error('Monitor: spike was not recorded')

        # Get data
        if not spikes:
            data = self.get('spike')
        else:
            if 'spike' in spikes.keys():
                data = spikes['spike']
            else:
                data = spikes


        # Compute the duration of the recordings
        duration = self._last_recorded_variables['spike']['stop'][-1] - self._last_recorded_variables['spike']['start'][-1]

        # Number of neurons
        neurons = self.object.ranks if isinstance(self.object, PopulationView) else range(self.object.size)

        # Compute fr
        fr = 0
        for neuron in neurons:
            fr += len(data[neuron])

        return fr/float(len(neurons))/duration/self.dt*1000.0



    def smoothed_rate(self, spikes:dict=None, smooth:float=0.) -> np.ndarray:
        """
        Computes the smoothed firing rate of the recorded spiking neurons.

        The first axis is the neuron index, the second is time.

        ```python
        m = net.monitor(pop, 'spike')
        net.simulate(1000.0)

        spikes = m.get('spike')
        r = m.smoothed_rate(spikes, smooth=100.)
        ```

        :param spikes: the dictionary of spikes returned by ``get('spike')``. If left empty, ``get('spike')`` will be called. Beware: this erases the data from memory.
        :param smooth: smoothing time constant. Default: 0.0 (no smoothing).

        """
        if not 'spike' in self._variables:
            Messages._error('Monitor: spike was not recorded')

        # Get data
        if not spikes:
            data = self.get('spike')
        else:
            if 'spike' in spikes.keys():
                data = spikes['spike']
            else:
                data = spikes

        import ANNarchy.cython_ext.Transformations as Transformations
        return Transformations.smoothed_rate(
            {
                'data': data,
                'start': self._last_recorded_variables['spike']['start'][-1],
                'stop': self._last_recorded_variables['spike']['stop'][-1]
            },
            smooth
        )

    def population_rate(self, spikes:dict=None, smooth:float=0.) -> np.ndarray:
        """
        Computes a smoothed firing rate for the population of recorded neurons.

        This method is faster than calling ``smoothed_rate`` and then averaging.

        If ``spikes`` is left empty, ``get('spike')`` will be called. Beware: this erases the data from memory.

        Example:

        ```python
        m = net.monitor(P[:1000], 'spike')
        net.simulate(1000.0)

        spikes = m.get('spike')
        r = m.population_rate(spikes, smooth=100.)
        ```

        :param spikes: the dictionary of spikes returned by ``get('spike')``.
        :param smooth: smoothing time constant. Default: 0.0 (no smoothing).

        """
        if not 'spike' in self._variables:
            Messages._error('Monitor: spike was not recorded')

        # Get data
        if not spikes:
            data = self.get('spike')
        else:
            if 'spike' in spikes.keys():
                data = spikes['spike']
            else:
                data = spikes

        import ANNarchy.cython_ext.Transformations as Transformations
        return Transformations.population_rate(
            {
                'data': data,
                'start': self._last_recorded_variables['spike']['start'][-1],
                'stop': self._last_recorded_variables['spike']['stop'][-1]
            },
            smooth
        )


def get_size(obj, seen=None):
    """
    Recursively determines the size of objects
    """
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size

class MemoryStats :
    """
    Create memory statistics for the main objects in ANNarchy. The current implementation
    focusses on the C++ simulation core. But this module could be further extended to measure
    also the Python objects.
    """
    def __init__(self):
        pass

    def print_cpp(self, net_id=0):
        """
        Print memory consumption of CPP objects. The method calls
        the size_in_bytes() methods implemented by the C++ modules.
        """
        for pop in NetworkManager().get_network(net_id=net_id).get_populations():
            if hasattr(pop, 'size_in_bytes'):
                print(pop.name, ":", self._human_readable_bytes(pop._size_in_bytes()))
            else:
                Messages._warning("MemoryStats.print_cpp(): the object", pop, "does not have a size_in_bytes() function.")

        for proj in NetworkManager().get_network(net_id=net_id).get_projections():
            if hasattr(proj, 'size_in_bytes'):
                print(proj.pre.name, "->", proj.post.name, "(", proj.target, "):", self._human_readable_bytes(proj._size_in_bytes()))
            else:
                Messages._warning("MemoryStats.print_cpp(): the object", proj, "does not have a size_in_bytes() function.")

        for mon in NetworkManager().get_network(net_id=net_id).get_monitors():
            if hasattr(proj, 'size_in_bytes'):
                print("Monitor on", mon.object.name, ":", self._human_readable_bytes(mon._size_in_bytes()))
            else:
                Messages._warning("MemoryStats.print_cpp(): the object", mon, "does not have a size_in_bytes() function.")

    def _human_readable_bytes(self, num):
        """
        All cpp functions return there size in bytes *num* as long int. This function
        divides this by 1024 until the result is lower than the next unit.
        """
        for x in ['bytes','KB','MB','GB']:
            if num < 1024.0:
                return "%3.2f %s" % (num, x)
            num /= 1024.0
        return "%3.1f%s" % (num, 'TB')


######################
# Static methods to plot spike patterns without a Monitor (e.g. offline)
######################
def raster_plot(spikes:dict, dt:float=1.0) -> tuple:
    """
    Returns two vectors representing for each recorded spike 1) the spike times and 2) the ranks of the neurons.

    :param spikes: the dictionary of spikes returned by ``get('spike')``.
    :param dt: the time step size in ms (default: 1.0ms).
    """
    times = []
    ranks=[]

    # Compute raster
    for n in spikes.keys():
        for t in spikes[n]:
            times.append(t)
            ranks.append(n)

    return dt * np.array(times), np.array(ranks)


def histogram(spikes:dict, bins:float=None, per_neuron:bool=False, recording_window:tuple=None, dt:float=1.0):
    """
    Returns a histogram for the recorded spikes in the population.

    :param spikes: the dictionary of spikes returned by ``get('spike')``.
    :param bins: the bin size in ms (default: dt).
    """
    if bins is None:
        bins =  dt

    bin_step = int(bins/dt)

    # Compute the duration of the recordings
    t_maxes = []
    t_mines = []
    for neuron in spikes.keys():
        if len(spikes[neuron]) == 0:
            continue
        t_maxes.append(np.max(spikes[neuron]))
        t_mines.append(np.min(spikes[neuron]))

    if recording_window is None:
        t_max = np.max(t_maxes)
        t_min = np.min(t_mines)
    else:
        t_min = recording_window[0]
        t_max = recording_window[1]
    duration = t_max - t_min

    # Number of bins
    nb_bins = int(duration/bin_step)
    #print(t_min, t_max, duration, nb_bins)

    if per_neuron:
        max_rank = np.amax([x for x in spikes.keys()])+1
        # Initialize histogram
        histo = [ [0 for _ in range(nb_bins+1)] for _ in range(max_rank) ]

        # Compute per step histogram
        for neuron in spikes.keys():
            for t in spikes[neuron]:
                histo[neuron][int((t-t_min)/float(bin_step))] += 1

    else:
        # Initialize histogram
        histo = [0 for t in range(nb_bins+1)]

        # Compute per step histogram
        for neuron in spikes.keys():
            for t in spikes[neuron]:
                histo[int((t-t_min)/float(bin_step))] += 1

    return np.array(histo)

def inter_spike_interval(spikes:dict, ranks:list=None, per_neuron:bool=False, dt:float=1.0):
    """
    Computes the inter-spike interval (ISI) for the recorded spike events of a population.

    :param spikes: the dictionary of spikes returned by ``get('spike')``.
    :param ranks: list of ranks.
    :param per_neuron: if True, the ISI will be computed per neuron, not globally.

    """
    isi = {}
    for neuron_rank, spike_events in spikes.items():
        # ISI computation requires at least 2 events
        if len(spike_events) < 2:
            continue

        # suppress unwanted neurons
        if ranks is not None:
            if neuron_rank not in ranks:
                continue

        # compute time difference between spike events
        tmp_isi=[]
        for idx in range(len(spike_events)-1):
            tmp_isi.append((spike_events[idx+1]-spike_events[idx])*dt)

        isi[neuron_rank] = tmp_isi

    if per_neuron:
        return isi
    else:
        res = []
        for val in isi.values():
            res.extend(val)
        return res

def coefficient_of_variation(spikes:dict, ranks:list=None, per_neuron:bool=False, dt=1.0):
    """
    Computes the coefficient of variation of the inter-spike intervals for the recorded spike events of a population.

    :param spikes: the dictionary of spikes returned by ``get('spike')``.
    :param ranks: list of ranks.
    :param per_neuron: if True, the ISI will be computed per neuron, not globally.
    """
    isi_per_neuron = inter_spike_interval(spikes, ranks=ranks, per_neuron=True, dt=dt)
    isi_cv = {}
    for neuron_rank, values in isi_per_neuron.items():
        if len(values) < 2:
            continue     # no meaningful mean/std possible

        # suppress unwanted neurons
        if ranks is not None:
            if neuron_rank not in ranks:
                continue

        isi_cv[neuron_rank] = np.std(values) / np.mean(values)

    if per_neuron:
        return isi_cv
    else:
        res = []
        for val in isi_cv.values():
            res.append(val)
        return res

def population_rate(spikes:dict, smooth:float=0.0):
    """
    Takes the recorded spikes of a population and returns a smoothed firing rate for the population of recorded neurons.

    This method is faster than calling ``smoothed_rate`` and then averaging.

    The first axis is the neuron index, the second is time.

    :param spikes: the dictionary of spikes returned by ``get('spike')``.
    :param smooth: smoothing time constant. Default: 0.0 (no smoothing).
    """
    # Compute the duration of the recordings
    t_maxes = []
    t_mines = []
    for neuron in spikes.keys():
        if len(spikes[neuron]) == 0 : continue
        t_maxes.append(np.max(spikes[neuron]))
        t_mines.append(np.min(spikes[neuron]))

    t_max = np.max(t_maxes)
    t_min = np.min(t_mines)

    import ANNarchy.cython_ext.Transformations as Transformations
    return Transformations.population_rate(
        {
            'data': spikes,
            'start':t_min,
            'stop': t_max
        },
        smooth
    )

def smoothed_rate(spikes:dict, smooth:float=0.):
    """
    Computes the smoothed firing rate of the recorded spiking neurons.

    The first axis is the neuron index, the second is time.

    :param spikes: the dictionary of spikes returned by ``get('spike')``. If left empty, ``get('spike')`` will be called. Beware: this erases the data from memory.
    :param smooth: smoothing time constant. Default: 0.0 (no smoothing).
    """
    # Compute the duration of the recordings
    t_maxes = []
    t_mines = []
    for neuron in spikes.keys():
        if len(spikes[neuron]) == 0 : continue
        t_maxes.append(np.max(spikes[neuron]))
        t_mines.append(np.min(spikes[neuron]))

    t_max = np.max(t_maxes)
    t_min = np.min(t_mines)

    import ANNarchy.cython_ext.Transformations as Transformations
    return Transformations.smoothed_rate(
        {
            'data': spikes,
            'start': t_min,
            'stop': t_max
        },
        smooth
    )

def mean_fr(spikes, duration=None, dt=1.0):
    """
    Computes the mean firing rate in the population during the recordings.

    :param spikes: the dictionary of spikes returned by ``get('spike')``.
    :param duration: duration of the recordings. By default, the mean firing rate is computed between the first and last spikes of the recordings.


    """
    if duration is None:

        # Compute the duration of the recordings
        t_maxes = []
        t_mines = []
        for neuron in spikes.keys():
            if len(spikes[neuron]) == 0 : continue
            t_maxes.append(np.max(spikes[neuron]))
            t_mines.append(np.min(spikes[neuron]))

        t_max = np.max(t_maxes)
        t_min = np.min(t_mines)
        duration = t_max - t_min

    nb_neurons = len(spikes.keys())

    # Compute fr
    fr = 0
    for neuron in spikes:
        fr += len(spikes[neuron])

    return fr/float(nb_neurons)/duration/dt*1000.0



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

profile_base_template = """#pragma once

#include <memory>
#include <iostream>
#include <string>
#include <vector>
#include <map>
#include <ostream>
#include <fstream>
#include <iomanip>
#include <math.h>
%(timer_import)s

#ifdef _DEBUG
    #define debug_cout(x) (std::cout << x << std::endl)
#else
    #define debug_cout(x)
#endif

%(measurement_class)s

/**
 *  @brief      Profiling class
 *  @details    Creation and initialization of the class are not thread-safe. 
 *              The class is implemented as singleton ensure uniqueness during
 *              runtime.
 */
extern class Profiling* prof_ptr;
class Profiling {
    std::vector<Measurement*> _datasets;                                ///< Instances of measurement objects. The index for correct access is retrieved from Profiling::_identifier
    std::map<std::pair<std::string, std::string>, int > _identifier;    ///< maps a (obj, func) descriptor to index for Profiling::_datasets
    std::map<std::string, int> _obj_to_ids;                             ///< needed for profile application

    std::ofstream _out_file;
%(timer_start_decl)s

public:

    /**
     *  @brief  Constructor
     */
    Profiling() {
        debug_cout("Create Profiling instance.");

        // HACK: the object constructor is now called by nanobind, need to update reference in C++ library
        prof_ptr = this;

%(timer_init)s

        if (false) {
            _out_file.open("results_%(config)s.xml", std::ofstream::out | std::ofstream::trunc);
            _out_file << "<root>" << std::endl;

        %(config_xml)s
        }
    }

    /**
     *  @brief  Destructor
     */
    ~Profiling() {
        debug_cout("Destroy Profiling instance.");

        for(auto it = _datasets.begin(); it != _datasets.end(); it++ )
            delete *it;

        _out_file << "</root>" << std::endl;
        _out_file.close();
    }

    /**
     *  @brief      Add a function to measurement dataset
     *  @details    A measurment is uniquely described by an object (either population or projection name)
     *              and a function name. The provided items will be used as key for the internal data map.
     *  @param[IN]  obj type as string (either pop, proj or net)
     *  @param[IN]  obj object name as string
     *  @param[IN]  func function name as string
     *  @return     Instance of measurement class. If the function is called multiple times no additional 
     *              object will be created.
     */
    Measurement* register_function(std::string type, std::string obj, int obj_id, std::string func, std::string label) {
        auto pair = std::pair<std::string, std::string>(obj, func);

        if ( _identifier.count(pair) == 0) { // not in list
            _identifier.insert(std::pair< std::pair<std::string, std::string>, int >(pair, _datasets.size()));
            _datasets.push_back(new Measurement(type, label));

            debug_cout( "(" + pair.first + ", " + pair.second + ") added to dataset." );
        }

        _obj_to_ids[obj] = obj_id;
        return get_measurement(obj, func);
    }

    /**
     *  @brief      Retrieve registered function from measurement dataset
     *  @details    A measurment is uniquely described by an object (either population or projection name)
     *              and a function name. The provided items will be used as key for the internal data map.
     *  @param[IN]  obj object name as string
     *  @param[IN]  func function name as string
     *  @return     Instance of measurement class. If the measurment were not registered before, no additional 
     *              object will be created. The function returns a 'nullptr' in this case.
     */
    Measurement* get_measurement(const std::string &obj, const std::string &func) {

        auto idx = _identifier.find(std::pair<std::string, std::string>(obj, func));
        if ( idx == _identifier.end() )
            return nullptr;

        return _datasets[idx->second];
    }

    /**
     *  @brief      Clear performance data of registered functions.
     *  @details    A measurment is uniquely described by an object (either population or projection name)
     *              and a function name. We iterate over all datasets and clear them seperatly.
     */
    void reset() {
        for( auto it = _datasets.begin(); it != _datasets.end(); it++ )
            (*it)->reset();
    }

    /**
     *  @brief      Evaluate registered functions.
     *  @details    A measurment is uniquely described by an object (either population or projection name)
     *              and a function name. We iterate over all datasets and evaluate them seperatly.
     */
    void evaluate() {
        for( auto it = _datasets.begin(); it != _datasets.end(); it++ )
            (*it)->evaluate();
    }

    void store() {
        //
        // To ensure, that the network related datasets are written first, we run two times
        // across the list ...
        //
        for( auto it = _identifier.begin(); it != _identifier.end(); it++ ) {
            if ( _datasets[it->second]->_raw_data.empty() || (_datasets[it->second]->_type.compare("net")!=0 ) )
                continue;    // nothing recorded, omit dataset

            _out_file << "  <dataset>" << std::endl;
            _out_file << "    <obj_type>" << _datasets[it->second]->_type << "</obj_type>" << std::endl;
            _out_file << "    <name>" << it->first.first << "</name>" << std::endl;
            _out_file << "    <func>" << it->first.second << "</func>" << std::endl;
            _out_file << "    <label>" << it->first.second << "</label>" << std::endl; // for top level objects "label" is equal to "func"
            _out_file << "    <mean>" << std::fixed << std::setprecision(4) << _datasets[it->second]->_mean << "</mean>"<< std::endl;
            _out_file << "    <std>" << std::fixed << std::setprecision(4) << _datasets[it->second]->_std << "</std>"<< std::endl;

            _out_file << "    <raw_data>";
            for(auto it2 = _datasets[it->second]->_raw_data.begin(); it2 != _datasets[it->second]->_raw_data.end(); it2++)
                _out_file << std::fixed << std::setprecision(4) << *it2 << " ";
            _out_file << "</raw_data>" << std::endl;

            _out_file << "  </dataset>" << std::endl;
        }

        for( auto it = _identifier.begin(); it != _identifier.end(); it++ ) {
            if ( _datasets[it->second]->_raw_data.empty() || (_datasets[it->second]->_type.compare("net")==0 ) )
                continue;    // nothing recorded, omit dataset

            _out_file << "  <dataset>" << std::endl;
            _out_file << "    <obj_type>" << _datasets[it->second]->_type << "</obj_type>" << std::endl;
            _out_file << "    <name>" << it->first.first << "</name>" << std::endl;
            _out_file << "    <func>" << it->first.second << "</func>" << std::endl;
            _out_file << "    <label>" << _datasets[it->second]->_label << "</label>" << std::endl;
            _out_file << "    <mean>" << std::fixed << std::setprecision(4) << _datasets[it->second]->_mean << "</mean>"<< std::endl;
            _out_file << "    <std>" << std::fixed << std::setprecision(4) << _datasets[it->second]->_std << "</std>"<< std::endl;

            _out_file << "    <raw_data>";
            for(auto it2 = _datasets[it->second]->_raw_data.begin(); it2 != _datasets[it->second]->_raw_data.end(); it2++)
                _out_file << std::fixed << std::setprecision(4) << *it2 << " ";
            _out_file << "</raw_data>" << std::endl;

            _out_file << "  </dataset>" << std::endl;
        }
    }

    /*
     *  Exported to Python
     */
    double get_avg_time(std::string obj, std::string func) {
        auto dataset = get_measurement(obj, func);
        if (dataset == nullptr)
            return 0.0;

        return dataset->_mean;
    }

    double get_std_time(std::string obj, std::string func) {
        auto dataset = get_measurement(obj, func);
        if (dataset == nullptr)
            return 0.0;

        return dataset->_std;
    }

    friend std::ostream& operator << (std::ostream& stream, const Profiling& profiling);
};

// out stream operators
inline std::ostream& operator << (std::ostream& stream, const Measurement& measure) {
    if ( measure._raw_data.empty() )
        return stream;

    stream << "mean: " << measure._mean << " ms, "
           << "std: " << measure._std << " ( over " << measure._raw_data.size() << " measurements )";

    return stream;
}

inline std::ostream& operator << (std::ostream& stream, const Profiling& profiling) {
    if ( profiling._datasets.empty() )
        return stream;

    for ( auto it = profiling._identifier.begin(); it != profiling._identifier.end(); it++ ) {
        stream << "  (" << it->first.first << ", " << it->first.second << ") - " << *(profiling._datasets[ it->second ]) << std::endl;
    }

    return stream;
}
"""

cpp11_profile_header = """
class Measurement{
    std::string _type;
    std::string _label;
    std::vector<double> _raw_data;

    double _mean;
    double _std;
    std::chrono::time_point<std::chrono::steady_clock> _start;
    std::chrono::time_point<std::chrono::steady_clock> _stop;

public:
    Measurement(std::string type, std::string label) {
        debug_cout("Create Measurement object");
        _type = type;
        _label = label;
        _mean = 0.0;
        _std = 0.0;
        _raw_data = std::vector<double>();
    }

    ~Measurement() {
        debug_cout("Destroy Measurement object");
        _raw_data.clear();
    }

    inline void start_wall_time() {
        _start = std::chrono::steady_clock::now();
    }

    inline void stop_wall_time() {
        _stop = std::chrono::steady_clock::now();
        std::chrono::duration<double, std::milli> dur = _stop - _start;
        // The measured times are stored in seconds
        _raw_data.push_back(std::chrono::duration<double>(dur).count());
    }

    void reset() {
        debug_cout("Reset Measurement object");
        _raw_data.clear();
        _mean = 0.0;
        _std = 0.0;
    }

    void evaluate() {
        if (_raw_data.empty())
            return;

        double sum = 0.0;
        int num_elem = _raw_data.size();

        // mean over all times
        for( auto it = _raw_data.begin(); it != _raw_data.end(); it++ ) {
            sum += *it;
        }

        _mean = double(sum) / double(num_elem);

        // variance over all times
        double var = 0;
        for( auto it = _raw_data.begin(); it != _raw_data.end(); it++ ) {
            var += (*it-_mean)*(*it-_mean);
        }

        _std = sqrt(var/double(num_elem));
    }

    friend std::ostream& operator << (std::ostream& stream, const Measurement& measure);
    friend class Profiling;
};
"""

cpp11_profile_template = {
    'include': """// Profiling
#include "Profiling.hpp"
extern class Profiling* prof_ptr;
""",
    'init': """
    //initialize profiler, create singleton instance
    prof_ptr->register_function("net", "network", 0, "step", "overall");
    prof_ptr->register_function("net", "network", 0, "psp", "overall");
    prof_ptr->register_function("net", "network", 0, "proj_step", "overall");
    prof_ptr->register_function("net", "network", 0, "proj_post_event", "overall");
    prof_ptr->register_function("net", "network", 0, "neur_step", "overall");
    prof_ptr->register_function("net", "network", 0, "global_op", "overall");
    prof_ptr->register_function("net", "network", 0, "record", "overall");
    prof_ptr->register_function("net", "network", 0, "rng", "overall");
    """,
    # Operations
    'proj_psp_pre': """// measure synaptic transmission
    auto measure_psp = prof_ptr->get_measurement("network", "psp");
    measure_psp->start_wall_time();
    """,
    'proj_psp_post': """// done
    measure_psp->stop_wall_time();
    """,
    'proj_step_pre': """// measure synaptic update
    auto measure_proj_step = prof_ptr->get_measurement("network", "proj_step");
    measure_proj_step->start_wall_time();
    """,
    'proj_step_post': """// done
    measure_proj_step->stop_wall_time();
    """,
    'proj_post_event_pre': """// measure post-event update
    auto measure_proj_post_event_step = prof_ptr->get_measurement("network", "proj_post_event");
    measure_proj_step->start_wall_time();
    """,
    'proj_post_event_post': """// done
    measure_proj_post_event_step->stop_wall_time();
    """,
    'neur_step_pre': """// measure population update
    auto measure_neur_step = prof_ptr->get_measurement("network", "neur_step");
    measure_neur_step->start_wall_time();
    """,
    'neur_step_post': """// done
    measure_neur_step->stop_wall_time();
    """,
    # Global operations
    'global_op_pre': """// measure global operations
    auto measure_global_op = prof_ptr->get_measurement("network", "global_op");
    measure_global_op->start_wall_time();
    """,
    'global_op_post': """// done
    measure_global_op->stop_wall_time();
    """,
    # Record
    'record_pre': """// measure record
    auto measure_rec = prof_ptr->get_measurement("network", "record");
    measure_rec->start_wall_time();
    """,
    'record_post': """// done
    measure_rec->stop_wall_time();
    """,
    # RNG
    'rng_pre': """// measure update rng
    auto measure_rng = prof_ptr->get_measurement("network", "rng");
    measure_rng->start_wall_time();
    """,
    'rng_post': """// done
    measure_rng->stop_wall_time();
    """,
    # Overall and setup
    'step_pre': """// before
    auto measure = prof_ptr->get_measurement("network", "step");
    measure->start_wall_time();
    """,
    'step_post': """// after
    measure->stop_wall_time();
    """,
    'run_pre': """
    prof_ptr->reset();
    """,
    'run_post': """
    prof_ptr->evaluate();
    if (false) {
        prof_ptr->store();
        std::cout << "profiling results: " << std::endl;
        std::cout << *prof_ptr << std::endl;
    }
    """,

    #
    # Execute the profile in each Object (i. e. populations, projections)
    'compute_psp': {
        'before' : "measure_psp->start_wall_time();",
        'after' : "measure_psp->stop_wall_time();"
    },
    'update_synapse': {
        'before' : "measure_step->start_wall_time();",
        'after' : "measure_step->stop_wall_time();"
    },
    'post_event': {
        'before' : "measure_pe->start_wall_time();",
        'after' : "measure_pe->stop_wall_time();"
    },
    'update_neuron': {
        'before' : "measure_step->start_wall_time();",
        'after' : "measure_step->stop_wall_time();"
    },
    'update_rng':{
        'before' : "measure_rng->start_wall_time();",
        'after' : "measure_rng->stop_wall_time();"
    },
    'update_delay':{
        'before' : "measure_delay->start_wall_time();",
        'after' : "measure_delay->stop_wall_time();"
    },
    'spike_gather': {
        'before' : "measure_sc->start_wall_time();",
        'after' : "measure_sc->stop_wall_time();"
    },
    'spike_prop': {
        'before' : "measure_prop->start_wall_time();",
        'after' : "measure_prop->stop_wall_time();"
    }
}

cpp11_omp_profile_template = {
    'include': """// Profiling
#include "Profiling.hpp"
extern class Profiling* prof_ptr;
""",
    'init': """
    //initialize profiler, create singleton instance
    prof_ptr->register_function("net", "network", 0, "step", "overall");
    prof_ptr->register_function("net", "network", 0, "psp", "overall");
    prof_ptr->register_function("net", "network", 0, "proj_step", "overall");
    prof_ptr->register_function("net", "network", 0, "proj_post_event", "overall");    
    prof_ptr->register_function("net", "network", 0, "neur_step", "overall");
    prof_ptr->register_function("net", "network", 0, "global_op", "overall");
    prof_ptr->register_function("net", "network", 0, "record", "overall");
    prof_ptr->register_function("net", "network", 0, "rng", "overall");
    """,
    # Operations
    'proj_psp_pre': """// measure synaptic transmission
    auto measure_psp = prof_ptr->get_measurement("network", "psp");
    #pragma omp barrier
    #pragma omp master
    measure_psp->start_wall_time();
    """,
    'proj_psp_post': """// done
    #pragma omp barrier
    #pragma omp master
    measure_psp->stop_wall_time();
    """,
    'proj_step_pre': """// measure synaptic update
    auto measure_proj_step = prof_ptr->get_measurement("network", "proj_step");
    #pragma omp barrier
    #pragma omp master
    measure_proj_step->start_wall_time();
    """,
    'proj_step_post': """// done
    #pragma omp barrier
    #pragma omp master
    measure_proj_step->stop_wall_time();
    """,
    'proj_post_event_pre': """// measure post-event update
    auto measure_proj_post_event_step = prof_ptr->get_measurement("network", "proj_post_event");
    #pragma omp barrier
    #pragma omp master
    measure_proj_step->start_wall_time();
    """,
    'proj_post_event_post': """// done
    #pragma omp barrier
    #pragma omp master
    measure_proj_post_event_step->stop_wall_time();
    """,
    'neur_step_pre': """// measure population update
    auto measure_neur_step = prof_ptr->get_measurement("network", "neur_step");
    #pragma omp barrier
    #pragma omp master
    measure_neur_step->start_wall_time();
    """,
    'neur_step_post': """// done
    #pragma omp barrier
    #pragma omp master
    measure_neur_step->stop_wall_time();
    """,
    # Global operations
    'global_op_pre': """// measure global operations
    auto measure_global_op = prof_ptr->get_measurement("network", "global_op");
    #pragma omp barrier
    #pragma omp master
    measure_global_op->start_wall_time();
    """,
    'global_op_post': """// done
    #pragma omp barrier
    #pragma omp master
    measure_global_op->stop_wall_time();
    """,
    # Record
    'record_pre': """// measure record
    auto measure_rec = prof_ptr->get_measurement("network", "record");
    #pragma omp barrier
    #pragma omp master
    measure_rec->start_wall_time();
    """,
    'record_post': """// done
    #pragma omp barrier
    #pragma omp master
    measure_rec->stop_wall_time();
    """,
    # RNG
    'rng_pre': """// measure update rng
    auto measure_rng = prof_ptr->get_measurement("network", "rng");
    #pragma omp master
    measure_rng->start_wall_time();
    """,
    'rng_post': """// done
    #pragma omp master
    measure_rng->stop_wall_time();
    """,

    # Overall and setup
    'step_pre': """// before
    auto measure = prof_ptr->get_measurement("network", "step");
    #pragma omp master
    measure->start_wall_time();
    """,
    'step_post': """// after
    #pragma omp master
    measure->stop_wall_time();
    """,
    'run_pre': """
    prof_ptr->reset();
    """,
    'run_post': """
    prof_ptr->evaluate();
    if (false) {
        prof_ptr->store();
        std::cout << "profiling results: " << std::endl;
        std::cout << *prof_ptr << std::endl;
    }
    """,

    #
    # Execute the profile in each Object (i. e. populations, projections)
    'compute_psp': {
        'before' : "#pragma omp barrier\n#pragma omp master\nmeasure_psp->start_wall_time();",
        'after' : "#pragma omp barrier\n#pragma omp master\nmeasure_psp->stop_wall_time();"
    },
    'update_synapse': {
        'before' : "#pragma omp barrier\n#pragma omp master\nmeasure_step->start_wall_time();",
        'after' : "#pragma omp barrier\n#pragma omp master\nmeasure_step->stop_wall_time();"
    },
    'post_event': {
        'before' : "#pragma omp barrier\n#pragma omp master\nmeasure_pe->start_wall_time();",
        'after' : "#pragma omp barrier\n#pragma omp master\nmeasure_pe->stop_wall_time();"
    },
    'update_neuron': {
        'before' : "#pragma omp barrier\n#pragma omp master\nmeasure_step->start_wall_time();",
        'after' : "#pragma omp barrier\n#pragma omp master\nmeasure_step->stop_wall_time();"
    },
    'update_rng':{
        'before' : "#pragma omp barrier\n#pragma omp master\nmeasure_rng->start_wall_time();",
        'after' : "#pragma omp barrier\n#pragma omp master\nmeasure_rng->stop_wall_time();"
    },
    'update_delay':{
        'before' : "#pragma omp barrier\n#pragma omp master\nmeasure_delay->start_wall_time();",
        'after' : "#pragma omp barrier\n#pragma omp master\nmeasure_delay->stop_wall_time();"
    },
    'spike_gather': {
        'before' : "#pragma omp barrier\n#pragma omp master\nmeasure_sc->start_wall_time();",
        'after' : "#pragma omp barrier\n#pragma omp master\nmeasure_sc->stop_wall_time();"
    },
    'spike_prop': {
        'before' : "#pragma omp barrier\n#pragma omp master\nmeasure_prop->start_wall_time();",
        'after' : "#pragma omp barrier\n#pragma omp master\nmeasure_prop->stop_wall_time();"
    }
}

cuda_profile_header = \
"""
class Measurement{
    std::string _type;
    std::string _label;
    std::vector<double> _raw_data;

    double _mean;
    double _std;
    cudaEvent_t _start;
    cudaEvent_t _stop;

public:
    Measurement(std::string type, std::string label) {
        debug_cout("Create Measurement object");
        _type = type;
        _label = label;
        _mean = 0.0;
        _std = 0.0;
        _raw_data = std::vector<double>();

        cudaEventCreate(&_start);
        cudaEventCreate(&_stop);
    }

    ~Measurement() {
        debug_cout("Destroy Measurement object");
        _raw_data.clear();

        cudaEventDestroy(_start);
        cudaEventDestroy(_stop);
    }

    inline void start_wall_time() {
        cudaEventRecord(_start);
    }

    inline void stop_wall_time() {
        cudaEventRecord(_stop);

        cudaEventSynchronize(_stop);
        float milliseconds = 0;
        cudaEventElapsedTime(&milliseconds, _start, _stop);

         // The measured times are stored in seconds
        _raw_data.push_back(double(milliseconds/1000.0));
    }

    void reset() {
        debug_cout("Reset Measurement object");
        _raw_data.clear();
        _mean = 0.0;
        _std = 0.0;
    }

    void evaluate() {
        if (_raw_data.empty())
            return;

        double sum = 0.0;
        int num_elem = _raw_data.size();

        // mean over all times
        for( auto it = _raw_data.begin(); it != _raw_data.end(); it++ ) {
            sum += *it;
        }

        _mean = double(sum) / double(num_elem);

        // variance over all times
        double var = 0;
        for( auto it = _raw_data.begin(); it != _raw_data.end(); it++ ) {
            var += (*it-_mean)*(*it-_mean);
        }

        _std = sqrt(var/double(num_elem));
    }

    friend std::ostream& operator << (std::ostream& stream, const Measurement& measure);
    friend class Profiling;
};
"""

cuda_profile_template = {
    'include': """// Profiling
#include "Profiling.hpp"
extern class Profiling* prof_ptr;
""",
    'init': """
    //initialize profiler, create singleton instance
    prof_ptr->register_function("net", "network", 0, "step", "overall");
    prof_ptr->register_function("net", "network", 0, "psp", "overall");
    prof_ptr->register_function("net", "network", 0, "proj_step", "overall");
    prof_ptr->register_function("net", "network", 0, "proj_post_event", "overall");
    prof_ptr->register_function("net", "network", 0, "neur_step", "overall");
    prof_ptr->register_function("net", "network", 0, "record", "overall");
    """,
    # Operations
    'proj_psp_pre': """// measure synaptic transmission
    auto measure_psp = prof_ptr->get_measurement("network", "psp");
    measure_psp->start_wall_time();
    """,
    'proj_psp_post': """// done
    measure_psp->stop_wall_time();
    """,
    'proj_step_pre': """// measure synaptic transmission
    auto measure_proj_step = prof_ptr->get_measurement("network", "proj_step");
    measure_proj_step->start_wall_time();
    """,
    'proj_step_post': """// done
    measure_proj_step->stop_wall_time();
    """,
    'neur_step_pre': """// measure population update
    auto measure_neur_step = prof_ptr->get_measurement("network", "neur_step");
    measure_neur_step->start_wall_time();
    """,
    'neur_step_post': """// done
    measure_neur_step->stop_wall_time();
    """,

    # Overall and setup
    'step_pre': """// before
    auto measure = prof_ptr->get_measurement("network", "step");
    measure->start_wall_time();
    """,
    'step_post': """// after
    measure->stop_wall_time();
    """,
    'run_pre': """
    prof_ptr->reset();
    """,
    'run_post': """
    prof_ptr->evaluate();
    if (false) {
        prof_ptr->store();
        std::cout << "profiling results: " << std::endl;
        std::cout << *prof_ptr << std::endl;
    }
    """,
    # Record
    'record_pre': """// measure record
    auto measure_rec = prof_ptr->get_measurement("network", "record");
    measure_rec->start_wall_time();
    """,
    'record_post': """// done
    measure_rec->stop_wall_time();
    """,
    #
    # Operations
    'compute_psp': {
        'before' : "proj%(id)s->measure_psp->start_wall_time();",
        'after' : "proj%(id)s->measure_psp->stop_wall_time();"
    },
    'update_synapse': {
        'before' : "proj%(id)s->measure_step->start_wall_time();",
        'after' : "proj%(id)s->measure_step->stop_wall_time();"
    },
    'post_event': {
        'before' : "proj%(id)s->measure_pe->start_wall_time();",
        'after' : "proj%(id)s->measure_pe->stop_wall_time();"
    },
    'update_neuron': {
        'before' : "pop%(id)s->measure_step->start_wall_time();",
        'after' : "pop%(id)s->measure_step->stop_wall_time();"
    },
    'spike_gather': {
        'before' : "pop%(id)s->measure_gather->start_wall_time();",
        'after' : "pop%(id)s->measure_gather->stop_wall_time();"
    },
    'spike_prop': {
        'before' : "measure_prop->start_wall_time();",
        'after' : "measure_prop->stop_wall_time();"
    },
}



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""
import numpy as np
import time
import secrets
import inspect

from typing import List
from dataclasses import dataclass, field

from ANNarchy.core.Population import Population
from ANNarchy.core.PopulationView import PopulationView
from ANNarchy.core.Projection import Projection
from ANNarchy.core.Monitor import Monitor
from ANNarchy.core.Neuron import Neuron
from ANNarchy.core.Synapse import Synapse
from ANNarchy.core.Constant import Constant

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages

import ANNarchy.extensions.bold as bold
import ANNarchy.core.Global as Global
import ANNarchy.core.Simulate as Simulate
import ANNarchy.core.IO as IO
import ANNarchy.generator.Compiler as Compiler

# Meta class to avoid forcing the constructor 
class NetworkMeta(type):

    def __call__(cls, *args, **kwargs):
        
        # Extract the seed and dt from kwargs if provided
        seed = kwargs.pop('seed', None)
        dt = kwargs.pop('dt', None)

        # Create an instance without calling __init__
        instance = cls.__new__(cls)

        # Call the parent class's __init__ methods first
        if isinstance(instance, Network):
            Network.__init__(instance, dt=dt, seed=seed)

        # Call the child's __init__ method
        if hasattr(cls, '__init__'):
            cls.__init__(instance, *args, **kwargs)

        instance._init_args = args
        instance._init_kwargs = kwargs

        return instance


@dataclass
class NetworkData :
    "Container for all the data of the network."
    populations: List = field(default_factory=list)
    projections: List = field(default_factory=list)
    monitors: List = field(default_factory=list)
    extensions: List = field(default_factory=list)
    constants: List = field(default_factory=list)
    instance = None
    compiled:bool = False
    directory:str = None

class Network (metaclass=NetworkMeta):
    """
    A network creates the populations, projections and monitors, and controls the simulation.

    ```python
    net = ann.Network(dt=1.0, seed=42)
    ```

    To create a population:

    ```python
    pop = net.create(100, neuron=ann.Izhikevich)
    ```

    To connect two populations:

    ```python
    proj = net.connect(pre=pop1, post=pop2, target='exc', synapse=ann.STDP)
    ```

    To monitor a population or projection:

    ```python
    m = net.monitor(pop, ['spike', 'v'])
    ```

    To compile the network:

    ```python
    net.compile()
    ```

    To simulate for one second:

    ```python
    net.simulate(1000.)
    ```
    
    Refer to the manual for more functionalities. `dt` and `seed` must be passed with keywords.

    :param dt: step size in milliseconds.
    :param seed: seed for the random number generators.
    """

    def __init__(self, dt:float=None, seed:int=None):
        # Constructor should only be called once
        if hasattr(self, '_initialized'):
            return
        self._initialized = True

        # Object tracking
        Messages._debug("Instantiate Network object", self)

        # Register the network in the NetworkManager()
        self.id = NetworkManager().add_network(self)

        # Create the data structure to store populations and projections
        self._data = NetworkData()

        # Get the default config
        ConfigManager().register_network(self.id)

        # dt
        if dt is not None:
            self.dt = dt

        # Draw a value for seed if not provided by user
        if seed is None:
            seed = secrets.randbits(32)  # Generates a random 32-bit integer

        # Store the seed value
        self._set_config('seed', seed)

        # initialize one RNG instance
        self._default_rng = np.random.default_rng(self.seed)
        # Store RNG state for Network.reset(reseed=True)
        self._default_rng_state = self._default_rng.bit_generator.state

        # Callbacks
        self._callbacks = []
        self._callbacks_enabled = True

        # Import constants from the global level
        self._import_constants()


    def __del__(self):
        
        # Overridden destructor for two reasons:
        # 
        # a) track destruction of objects
        # b) manually deallocate C++ container data
        # 
        # Hint: this function can be called explicitly (which is not recommended in many cases) or as
        #       finalizer from the garbage collection. If called explicitely, one should take in mind,
        #       that the function will be called twice. The better approach is to trigger this function
        #       by del on the network object

        # clear attached objects
        self.clear()

        # de-register the network instance
        NetworkManager().remove_network(self)

    def create(
            self,
            geometry: tuple | int = None, 
            neuron: Neuron = None, 
            stop_condition:str = None, 
            name:str = None,
            population: Population = None,

            # Internal use only
            storage_order:str = 'post_to_pre', 
            ) -> Population:
        """
        Adds a population of neurons to the network.

        ```python
        net = ann.Network()
        pop = net.create(geometry=100, neuron=ann.Izhikevich, name="Excitatory population")
        ```

        Specific populations (e.g. `PoissonPopulation()`) can also be passed to the `population` argument, or simply as the first argument, in both cases do not provide other arguments.:

        ```python
        pop = net.create(population=ann.PoissonPopulation(100, rates=20.))
        # or
        pop = net.create(ann.PoissonPopulation(100, rates=20.))
        ```

        :param geometry: population geometry as tuple. If an integer is given, it is the size of the population. If an instance of `Population` is given, do not provide other arguments.
        :param neuron: `Neuron` instance. It can be user-defined or a built-in model.
        :param name: unique name of the population (optional).
        :param stop_condition: a single condition on a neural variable which can stop the simulation whenever it is true.
        :param population: instance of a `Population`. If given, do not provide other arguments.
        """
        # if both population and geometry are None, error
        if geometry is None and population is None:
            Messages._error("Network.create(): either 'geometry' or 'population' argument must be provided.")
        
        # if both are given as population instances, error
        if isinstance(geometry, Population) and population is not None:
            Messages._error("Network.create(): cannot provide both 'geometry' and 'population' as Population instances.")
        
        # if population is given, check its type
        if population is not None and not isinstance(population, Population):
            Messages._error("Network.create(): 'population' argument only accepts instances of ann.Population and its subclasses.")
        
        # if a population instance is given, check that the other arguments are None
        if isinstance(geometry, Population) or population is not None:
            if neuron is not None or name is not None or stop_condition is not None:
                Messages._error("Network.create(): do not give other arguments when a `Population` instance is provided. Define them when creating the given `Population` instance.")
        if population is not None:
            if geometry is not None:
                Messages._error("Network.create(): do not give other arguments when a `Population` instance is provided. Define them when creating the given `Population` instance.")

        if isinstance(geometry, Population):  # trick if one does provide `Population` as first argument
            # Population is already created
            pop = geometry._copy(self.id)
        elif population is not None:
            # Population is already created
            pop = population._copy(self.id)
        else:
            # Create the population
            pop = Population(
                geometry=geometry, 
                neuron=neuron, 
                name=name, 
                stop_condition=stop_condition,
                storage_order=storage_order, 
                copied=False, 
                net_id=self.id
            )

        return pop
    
    def connect(
            self,
            pre: str | Population = None, 
            post: str | Population = None, 
            target: str = "", 
            synapse: Synapse = None, 
            name:str = None, 
            projection:Projection = None,
            # Internal
            disable_omp:bool = True, 
        ) -> "Projection":
        """
        Connects two populations by creating a projection.

        ```python
        net.connect(pre=pop1, post=pop2, target="exc", synapse=STDP)
        ```

        If the `synapse` argument is omitted, defaults synapses without plastivity will be used (`psp = "w * pre.r"` for rate-coded projections, `pre_spike="g_target += w"` for spiking ones.).

        Specific projections can be passed to the `projection` argument, or as the first unnamed argument. In both cases, do not provide other arguments.

        ```python
        net.connect(projection=ann.DecodingProjection(pre, post, 'exc'))
        ```

        :param pre: pre-synaptic population. If an instance of `Projection` is given, do not provide other arguments.
        :param post: post-synaptic population.
        :param target: type of the connection.
        :param synapse: `Synapse` class or instance.
        :param name: (optional) name of the Projection.
        :param projection: specific projection. If given, do not provide other arguments.
        """
        # if both projection and pre are None, error
        if pre is None and projection is None:
            Messages._error("Network.connect(): either 'pre' or 'projection' argument must be provided.")

        # if both are given as projection instances, error
        if isinstance(pre, Projection) and projection is not None:
            Messages._error("Network.connect(): cannot provide both 'pre' and 'projection' arguments.")

        # if projection is given, check its type
        if projection is not None and not isinstance(projection, Projection):
            Messages._error("Network.connect(projection=proj) only accepts instances of ann.Projection and its subclasses.")

        # if a projection instance is given, check that the other arguments are not provided
        if isinstance(pre, Projection) or projection is not None:
            if post is not None or target != "" or synapse is not None or name is not None:
                Messages._error("Network.connect(): do not give other arguments when a `Projection` instance is provided. Define them when creating the given `Projection` instance.")
        if projection is not None:
            if pre is not None:
                Messages._error("Network.connect(): do not give other arguments when a `Projection` instance is provided. Define them when creating the given `Projection` instance.")

        if isinstance(pre, Projection):  # trick if one does provide `Projection` as first argument
            # Projection is already created
            proj = pre._copy(pre.pre, pre.post, self.id)
        elif projection is not None:
            # Projection is already created
            proj = projection ._copy(projection.pre, projection.post, self.id)
        else:
            # Create the projection
            if post is None:
                Messages._error("Network.connect(): the post population must be provided.")
            if target == "":
                Messages._error("Network.connect(): the target must be specified.")

            proj = Projection(
                pre = pre, 
                post = post, 
                target = target, 
                synapse = synapse, 
                name = name, 
                # Internal
                disable_omp = disable_omp, 
                copied = False,
                net_id = self.id,
            )
        
        return proj
    
    def monitor(
            self,
            obj: Population | PopulationView | Projection, 
            variables:list=[], 
            period:float=None, 
            period_offset:float=None, 
            start:bool=True, 
            name:str=None,
            ) -> Monitor:
        """
        Creates a Monitor on variables of the specified object.

        The object can be an instance of either `Population`, `PopulationView`, `Dendrite` or `Projection`.

        The variables must be declared by the neuron or synapse type. For spiking neurons, `'spike'` can be also recorded.

        ```python
        m = net.monitor(pop, ['v', 'spike'])
        m = net.monitor(proj.dendrite(0), 'w')
        ```

        By default, the monitors start recording right after `Network.compile()`, but you can hold them with `start=False`. Starting the recordings necessitates then to call the ``start()``mehtod. Pausing/resuming the recordings is cheived through the `pause()` and `resume()`.

        ```python
        m.start() # Start recording
        net.simulate(T)
        m.pause() # Pause recording
        net.simulate(T)
        m.resume() # Resume recording
        net.simulate(T)

        data = m.get() # Get the data
        ```

        :param obj: object to monitor. Must be a `Population`, `PopulationView`, `Dendrite` or `Projection` object.
        :param variables: single variable name or list of variable names to record (default: []).
        :param period:
        :param period_offset:
        :param start:
        :param name:
        """
        
        if isinstance(obj, Monitor): # trick if one does not use obj=
            monitor = obj._copy(self.id)
        else:
            monitor = Monitor(
                obj=obj, 
                variables=variables, 
                period=period, 
                period_offset=period_offset, 
                start=start, 
                name=name,
                net_id=self.id
            )
            
        return monitor
    
    def boldmonitor(
            self,
            populations: list=None,
            bold_model: bold.BoldModel = None,
            mapping: dict={'I_CBF': 'r'},
            scale_factor: list[float]=None,
            normalize_input: list[int]=None,
            recorded_variables: list[str]=None,
            start:bool=False,
            ) -> "bold.BoldMonitor":
        """
        Monitors the BOLD signal of several populations using a computational model.

        The BOLD monitor transforms one or two input population variables (such as the mean firing rate) into a recordable BOLD signal according to a computational model (for example a variation of the Balloon model).

        Several models are available or can be created with a `bold.BoldModel` instance. These models must be explicitly imported:

        ```python
        import ANNarchy.extensions.bold as bold

        m_bold = net.boldmonitor(
            # Recorded populations
            populations = [pop1, pop2], 
            # BOLD model to use (default is balloon_RN)
            bold_model = bold.balloon_RN(), 
            # Mapping from pop.r to I_CBF
            mapping = {'I_CBF': 'r'}, 
            # Time window to compute the baseline.
            normalize_input = 2000,  
            # Variables to be recorded
            recorded_variables = ["I_CBF", "BOLD"]  
        )

        # Unlike regular monitors, BOLD monitors must be explicitly started
        m_bold.start()
        net.simulate(5000) 
        bold_data = m_bold.get("BOLD")
        ```

        :param populations: list of recorded populations.
        :param bold_model: computational model for BOLD signal defined as a BoldModel object. Default is `bold.balloon_RN`.
        :param mapping: mapping dictionary between the inputs of the BOLD model (`I_CBF` for single inputs, `I_CBF` and `I_CMRO2` for double inputs in the provided examples) and the variables of the input populations. By default, `{'I_CBF': 'r'}` maps the firing rate `r` of the input population(s) to the variable `I_CBF` of the BOLD model. 
        :param scale_factor: list of float values to allow a weighting of signals between populations. By default, the input signal is weighted by the ratio of the population size to all populations within the recorded region.
        :param normalize_input: list of integer values which represent a optional baseline per population. The input signals will require an additional normalization using a baseline value. A value different from 0 represents the time period for determing this baseline in milliseconds (biological time).
        :param recorded_variables: which variables of the BOLD model should be recorded? (by default, the output variable of the BOLD model is added, e.g. ["BOLD"] for the provided examples).
        :param start: whether to start recording directly.
        """

        if bold_model is None:
            bold_model = bold.balloon_RN

        boldmonitor = bold.BoldMonitor(
                populations=populations,
                bold_model=bold_model,
                mapping=mapping,
                scale_factor=scale_factor,
                normalize_input=normalize_input,
                recorded_variables=recorded_variables,
                start=start,
                net_id=self.id,
            )

        return boldmonitor
    
    def constant(self, name, value):
        """
        Adds a constant to the network.

        ```python
        c = net.constant('c', 2.0)

        # `c` can be used in a neuron/synapse definition
        neuron = ann.Neuron(equations="r = c * sum(exc)")

        # Change the value of the constant in the network.
        c.set(10.0) 
        ```

        :param name: name of the constant.
        :param value: initial value of the constant.
        """
        constant = Constant(name, value, net_id=self.id)
        return constant
    
    def _import_constants(self):
        """
        Tells the network to use the constants defined at the global level.
        """
        constants = NetworkManager().magic_network().get_constants()

        for c in constants:
            # Create the constant inside this network
            c._set_child_network(self.id)
            c._copy(self.id)
    
    ###################################
    # Compile
    ###################################

    def compile(self,
                directory:str='annarchy',
                clean:bool=False,
                compiler:str="default",
                compiler_flags:list[str]="default",
                add_sources:str="",
                extra_libs:str="",
                cuda_config:dict={'device': 0},
                annarchy_json:str="",
                silent:bool=False,
                debug_build:bool=False,
                profile_enabled:bool=False
        ) -> None:
        """
        Compiles the network.

        :param directory: name of the subdirectory where the code will be generated and compiled. Default: "./annarchy/".
        :param clean: boolean to specifying if the library should be recompiled entirely or only the changes since last compilation (default: False).
        :param compiler: C++ compiler to use. Default: g++ on GNU/Linux, clang++ on OS X. Valid compilers are [g++, clang++].
        :param compiler_flags: platform-specific flags to pass to the compiler. Default: "-march=native -O2". Warning: -O3 often generates slower code and can cause linking problems, so it is not recommended.
        :param cuda_config: dictionary defining the CUDA configuration for each population and projection.
        :param annarchy_json: compiler flags etc are stored in a .json file normally placed in the home directory. With this flag one can directly assign a file location.
        :param silent: defines if the "Compiling... OK" should be printed.

        """
        Compiler.compile(
            directory=directory, 
            clean=clean, 
            silent=silent, 
            debug_build=debug_build, 
            add_sources=add_sources, 
            extra_libs=extra_libs, 
            compiler=compiler, 
            compiler_flags=compiler_flags, 
            cuda_config=cuda_config, 
            annarchy_json=annarchy_json, 
            profile_enabled=profile_enabled, 
            net_id=self.id)
        
    
    ###################################
    # Simulation
    ###################################
    def simulate(self, duration:float, measure_time:bool=False):
        """
        Runs the network for the given duration in milliseconds. 
        
        The number of simulation steps is  computed relative to the discretization step `dt` declared in the constructor (default: 1 ms):

        ```python
        net.simulate(1000.0)
        ```

        :param duration: the duration in milliseconds.
        :param measure_time: defines whether the simulation time should be printed.

        """
        Simulate.simulate(duration, measure_time, net_id=self.id)

    def simulate_until(self, max_duration:float, population:"Population", operator:str='and', measure_time:bool=False) -> float:
        """
        Runs the network for the maximal duration in milliseconds until a `stop_condition` is met.
        
        Whenever the `stop_condition` defined in `population` becomes true, the simulation is stopped.

        The method returns the actual duration of the simulation in milliseconds.

        One can specify several populations. If the stop condition is true for any of the populations, the simulation will stop ('or' function).

        Example:

        ```python
        pop1 = net.create( ..., stop_condition = "r > 1.0 : any")
        
        net.compile()
        
        net.simulate_until(max_duration=1000.0. population=pop1)
        ```

        :param max_duration: the maximum duration of the simulation in milliseconds.
        :param population: the (list of) population whose ``stop_condition`` should be checked to stop the simulation.
        :param operator: operator to be used ('and' or 'or') when multiple populations are provided (default: 'and').
        :param measure_time: defines whether the simulation time should be printed (default=False).
        """
        return Simulate.simulate_until(max_duration=max_duration, population=population, operator=operator, measure_time=measure_time, net_id=self.id)

    def step(self) -> None:
        """
        Performs a single simulation step (duration = `dt`).
        """
        Simulate.step(self.id)

    def reset(self, populations:bool=True, projections:bool=False, synapses:bool=False, monitors:bool=True, reseed_rng:bool=True) -> None:
        """
        Reinitialises the network to its state before the call to `compile()`.  The network time will be set to 0ms.

        :param populations: if True (default), the neural parameters and variables will be reset to their initial value.
        :param projections: if True, the synaptic parameters and variables (except the connections) will be reset (default=False).
        :param synapses: if True, the synaptic weights will be erased and recreated (default=False).
        :param monitors: if True, the monitors will be emptied and reset (default=True).
        ;param reseed_rng: if True, RNG generators will be reset using the stored seed (default=True).
        """

        self.instance.set_time(0)

        if populations:
            for pop in self.get_populations():
                pop.reset()

            # pop.reset only clears spike container with no or uniform delay
            for proj in self.get_projections():
                if hasattr(proj.cyInstance, 'reset_ring_buffer'):
                    proj.cyInstance.reset_ring_buffer()

        if synapses and not projections:
            Messages._warning("reset(): if synapses is set to true this automatically enables projections==true")
            projections = True

        if projections:
            for proj in self.get_projections():
                proj.reset(attributes=-1, synapses=synapses)

        if monitors:
            for monitor in self.get_monitors():
                monitor.reset()

        if reseed_rng:
            # Python:   re-initialize the RNG with initially stored state
            self._default_rng.bit_generator.state = self._default_rng_state

            # CPP:      re-initialize the RNG with the stored configuration
            if self._get_config('disable_parallel_rng'):
                self.instance.set_seed(self._get_config('seed'), 1, self._get_config('use_seed_seq'))
            else:
                self.instance.set_seed(self._get_config('seed'), self._get_config('num_threads'), self._get_config('use_seed_seq'))

    def enable_learning(self, projections:list=None, period:float=None, offset:float=None) -> None:
        """
        Enables learning for all projections.

        :param projections: the projections whose learning should be enabled. By default, all the existing projections are disabled.
        """
        if not projections:
            projections = self._data.projections
        for proj in projections:
            proj.enable_learning(period=period, offset=offset)

    def disable_learning(self, projections:list=None) -> None:
        """
        Disables learning for all projections.

        :param projections: the projections whose learning should be disabled. By default, all the existing projections are disabled.
        """
        if not projections:
            projections = self._data.projections
        for proj in projections:
            proj.disable_learning()

        
    def clear(self):
        """
        Empties the network to prevent a memory leak until the garbage collector wakes up.
        """
        for pop in self._data.populations:
            pop._clear()
            del pop

        for proj in self._data.projections:
            proj._clear()
            del proj

        for mon in self._data.monitors:
            mon._clear()
            del mon

        for ext in self._data.extensions:
            ext._clear()
            del ext

        for const in self._data.constants:
            del const

    ###################################
    # Parallel run
    ###################################

    def _instantiate(self, import_id=-1):
        """
        Instantiates the network using the code generated for the network of id `import_id`.
        """
        # Instantiate the network (but do not compile it) as if it had the provided id.
        Compiler._instantiate(self.id, import_id=import_id, cuda_config=None, user_config=None, core_list=None)

    def parallel_run(
            self, 
            method,
            number:int, 
            max_processes:int=-1, 
            seeds: int | str| list[int] = None,
            measure_time:bool=False, 
            **kwargs
        ):
        """
        Runs the provided method for multiple copies of the network.

        **Important:** The network must have defined as a subclass of `Network`, not a `Network` instance where `create()` and `connect()` where sequentially called. 
        
        See the manual for more explanations:

        ```python
        class PulseNetwork(ann.Network):
            def __init__(self):
                self.create(...)

        # Simulation method
        def simulation(net, duration=1000.):
            net.simulate(duration)
            t, n = net.m.raster_plot()
            return t, n

        # Create a network
        net = PulseNetwork()
        net.compile()

        # Parallel simulation
        results = net.parallel_run(method=simulation, number=4)
        ```

        :param method: method invoked for each copy of the network.
        :param number: number of simulations. 
        :param max_processes: maximum number of concurrent processes. Defaults to the number of cores.
        :param seeds: list of seeds for each network. If `None`, the seeds will be be randomly set. If `'same'`, it will be the same as the current network. If `'sequential'`, the seeds will be incremented for each network (42, 43, 44, etc).
        :param measure_time: if the total duration of the simulation should be reported at the end.
        """
        Messages._debug("Network was created with ", self._init_args, "and", self._init_kwargs)

        if type(self) is Network:
            Messages._error("Network.parallel_run(): the network must be an instance of a class deriving from Network, not Network itself.")

        if measure_time:
            tstart = time.time()

        # Seed
        if seeds is None: # Random seed for each network
            seeds = [None for _ in range(number)]
        elif seeds == 'same': # Same seed as the current one for each network
            seeds = [self._get_config('seed') for _ in range(number)]
        elif seeds == 'sequential': # Sequential seeds
            seeds = [self._get_config('seed') + i + 1 for i in range(number)]

        seeds = list(seeds)
        if len(seeds) != number:
            Messages._error("Network.parallel_run(): the list of seeds must have the same size as the number of simulations.")

        # Retrieve dictionary of parameters passed to the current constructor
        current_init_args = self._retrieve_initargs()

        # Arguments to the instances and method
        init_args = [current_init_args.copy() for _ in range(number)]
        method_args = [{} for _ in range(number)]

        # Iterate over 
        for key, val in kwargs.items():
            if key in current_init_args.keys(): # put the value in the constructor
                if isinstance(val, list):
                    if len(val) != number:
                        Messages._error("Network.parallel_run(): the list of values for", key, "must have the same size as the number of simulations.")
                    for i in range(number): init_args[i][key] = val[i]
                else:
                    for i in range(number): init_args[i][key] = val

            else: # put it in the simulation method
                if isinstance(val, list):
                    if len(val) != number:
                        Messages._error("Network.parallel_run(): the list of values for", key, "must have the same size as the number of simulations.")
                    for i in range(number): method_args[i][key] = val[i]
                else:
                    for i in range(number): method_args[i][key] = val

        # Config
        config = ConfigManager().get_config(self.id)
        config.pop('dt', 0.0)
        config.pop('seed', 0)

        # Import multiprocessing here to avoid warnings when setting the start method
        import multiprocessing as mp

        # Create arguments
        args = []
        for i in range(number):
            parameters = (
                    self.id, 
                    self.__class__, 
                    config,
                    method,
                    self.dt,
                    seeds[i],
                    init_args[i], 
                    method_args[i],
            )
            args.append(parameters)

        # Create and run the processes
        with mp.Pool(processes=min(number, max_processes if max_processes > 0 else mp.cpu_count())) as pool:
            results = pool.map(
                        self._worker, # method to call
                        args # arguments
                    )

        # Time measurement
        if measure_time:
            Messages._print('Simulating', number, 'networks in parallel took', time.time() - tstart, 'seconds.')

        return results

    @staticmethod
    def _worker(params):
        "Worker method called by parallel_run()."

        # Get the parameters
        id, classname, config, method, dt, seed, init_args, method_args = params

        # Create an instance with the same parameters as the original instance
        net = classname(dt=dt, seed=seed, **init_args)

        # Set the config
        ConfigManager().set_config(net.id, config)

        # Transfer compilation state of "parent" network
        net.compiled = NetworkManager().get_network(id).compiled

        # Instantiate the network but not compile
        net._instantiate(import_id=id)
        
        # Run the simulation
        result = method(net, **method_args)
        
        # Delete the network
        net.__del__()
        
        return result
    
    def _retrieve_initargs(self):

        # args and kwargs were saved here
        Messages._debug(self._init_args, self._init_kwargs,)

        # inspect the signature of the constructor
        signature = inspect.signature(self.__init__)
        params = list(signature.parameters.values())
        
        # Create a dictionary to hold the reconstructed arguments
        reconstructed_args = {
            param.name: param.default
            for param in params
            if param.default is not param.empty
        }
        
        # Map *args to the corresponding parameter names
        for i, arg in enumerate(self._init_args):
            if i < len(params):
                param_name = params[i].name
                reconstructed_args[param_name] = arg
        
        # Add the **kwargs to the reconstructed arguments
        reconstructed_args.update(self._init_kwargs)
        
        return reconstructed_args
    

    def copy(self, *args, **kwargs):
        """
        Returns a new instance of the Network class, using the provided arguments to the constructor. 
        
        Beware, `Network.compile()` is not called, only the instantiation of the data structures. Nothing in the constructor should induce a recompilation.
        """
        if not 'dt' in kwargs.keys():
            kwargs.update(dict(dt=self.dt))

        # Sanity check
        if not self.compiled:
            Messages._error("The copied network should be compiled before calling Network.copy().")

        # Create an instance of the child class
        net = self.__class__(*args, **kwargs)

        # Transfer the compiled state from "parent" network
        net.compiled = self.compiled

        # Instantiate the network with the current id.
        net._instantiate(self.id)

        return net


    ###################################
    # Config
    ###################################
    def _get_config(self, key:str):
        "Returns the config value with the given key."
        return  ConfigManager().get(key=key, net_id=self.id)
    def _set_config(self, key:str, value):
        "Sets the config value with the given key."
        return  ConfigManager().set(key=key, value=value, net_id=self.id)
    
    def config(self, *args, **kwargs):
        """
        Configuration of the network. 

        This method is equivalent to calling `setup()` at the global level, but only influences the current network. The initial configuration of the network copies the values set in `setup()` at the time of the creation of the network.
        
        It can be called multiple times until `compile()` is called, new values of keys erasing older ones.

        The only functional difference with `setup()` is the seed, which should be passed to the constructor of `Network`, otherwise any random number generation in the constructor might be unseeded. `dt` can also be passed to the constructor, but setting it in `config()` is also fine.
        
        It takes various optional arguments. The most useful ones are:

        * `dt`: simulation step size in milliseconds (default: 1.0).
        * `paradigm`: parallel framework for code generation. Accepted values: "openmp" or "cuda" (default: "openmp").
        * `method`: default method to numerize the ODEs. Default is the explicit forward Euler method ('explicit').
        * `precision`: default floating precision for variables in ANNarchy. Accepted values: "float" or "double" (default: "double")
        * `structural_plasticity`: allows synapses to be dynamically added/removed during the simulation (default: False).
        * `seed`: the seed (integer) to be used in the random number generators (default = None is equivalent to time(NULL)).
        * `num_threads`: number of treads used by openMP (overrides the environment variable ``OMP_NUM_THREADS`` when set, default = None).

        Flags related to the optimization of the simulation kernels are:

        * `sparse_matrix_format`: the default matrix format for projections in ANNarchy (by default: List-In-List for CPUs and Compressed Sparse Row). Note that this affects only the C++ data structures.
        * `sparse_matrix_storage_order`: encodes whether the row in a connectivity matrix encodes pre-synaptic neurons (post_to_pre, default) or post-synaptic neurons (pre_to_post). Note that affects only the C++ data structures.
        * `only_int_idx_type`: if set to True (default) only signed integers are used to store pre-/post-synaptic ranks which was default until 4.7. If set to False, the index type used in a single projection is selected based on the size of the corresponding populations.
        * `visible_cores`: allows a fine-grained control which cores are useable for the created threads (default = [] for no limitation). It can be used to limit created openMP threads to a physical socket.

        The following parameters are mainly for debugging and profiling, and should be ignored by most users:

        * `verbose`: shows details about compilation process on console (by default False). Additional some information of the network construction will be shown.
        * `suppress_warnings`: if True, warnings (e. g. from the mathematical parser) are suppressed.
        * `show_time`: if True, initialization times are shown. Attention: verbose should be set to True additionally.
        * `disable_shared_library_time_offset`: by default False. If set to True, the shared library generated by ANNarchy will not be extended by time offset.
        """

        # RNG-related arguments are treated differently
        rng_changed = False
        if 'use_seed_seq' in kwargs.keys():
            seed_seq = kwargs.pop('use_seed_seq', self._get_config('use_seed_seq'))
            self._set_config('use_seed_seq', seed_seq)
            rng_changed = True
        if 'disable_parallel_rng' in kwargs.keys():
            par_rng = kwargs.pop('disable_parallel_rng', self._get_config('disable_parallel_rng'))
            self._set_config('disable_parallel_rng', par_rng)
            rng_changed = True
        if 'seed' in kwargs.keys():
            # update the value for both NumPy and C++ core (when the latter is already compiled)
            self.seed = int(kwargs.pop('seed'))
            rng_changed = True

        if rng_changed:
            Messages._debug("RNG generators are now configured as following:")
            Messages._debug("  seed={} (accounts for C++ and NumPy)".format(self.seed))
            Messages._debug("  use_seed_seq={}".format("True" if self._get_config('use_seed_seq') else "False"))
            Messages._debug("  disable_parallel_rng={}".format("True" if self._get_config('disable_parallel_rng') else "False"))

        # Process all other provided key-value pairs
        for key, value in kwargs.items():
            # sanity check: filter out performance flags
            if key in ConfigManager()._performance_related_config_keys:
                Messages._error("The performance related flag '"+key+"' can not be configured by setup()")
            elif key in ConfigManager().keys():
                ConfigManager().set(key, value, net_id=self.id)
            else:
                Messages._warning('setup(): unknown key:', key)

            if key == 'sparse_matrix_format':
                # check if this is a supported format
                if kwargs[key] not in ["lil", "csr", "csr_vector", "csr_scalar", "dense", "ell", "ellr", "sell", "coo", "bsr", "hyb", "auto"]:
                    Messages._error("The value", kwargs[key], "provided to sparse_matrix_format is not valid.")


    @property
    def seed(self) -> int:
        "Seed for the random number generator being used for both Python and C++."
        return self._get_config('seed')

    @seed.setter
    def seed(self, seed: int) -> None:
        "Prevent the setting of a seed by hand."
        Messages._error("The seed argument should not be overwritten.")

    @property
    def default_rng(self) -> np.random.Generator:
        """
        Get a pre-seeded RNG instance. During construction of the network object, a numpy.random.Generator
        object has been instantiated. This can be used to initialize random distributions. e.g., instances of
        ANNarchy.core.RandomDistribution.
        """
        return self._default_rng

    @default_rng.setter
    def default_rng(self, new_rng: np.random.Generator) -> None:
        "Prevent the setting of RNG instance by hand."
        Messages._error("The default_rng argument should not be overwritten.")

    @property
    def dt(self) -> float:
        "Step size in milliseconds for the integration of the ODEs."
        return self._get_config('dt')

    @dt.setter
    def dt(self, dt:float) -> None:
        self._set_config('dt', dt)



    ###################################
    # IO
    ###################################
    def load(self, filename:str, populations:bool=True, projections:bool=True, pickle_encoding:str=None):
        """
        Loads parameters and variables from a file created with `Network.save()`.

        :param filename: filename, may contain relative or absolute path.
        :param populations: if `True`, population data will be saved.
        :param projections: if `True`, projection data will be saved.
        :param pickle_encoding: optional parameter provided to the pickle.load() method. If set to None the default is used.
        """
        IO.load(filename=filename, populations=populations, projections=projections, pickle_encoding=pickle_encoding, net_id=self.id)

    def save(self, filename:str, populations:bool=True, projections:bool=True):
        """
        Saves the parameters and variables of the networkin a file.

        * If the extension is '.npz', the data will be saved and compressed using `np.savez_compressed` (recommended).
        * If the extension is '.mat', the data will be saved as a Matlab 7.2 file. Scipy must be installed.
        * If the extension ends with '.gz', the data will be pickled into a binary file and compressed using gzip.
        * Otherwise, the data will be pickled into a simple binary text file using cPickle.

        **Warning:** The '.mat' data will not be loadable by ANNarchy, it is only for external analysis purpose.

        Example:

        ```python
        net.save('results/init.npz')

        net.save('results/init.data')

        net.save('results/init.txt.gz')

        net.save('1000_trials.mat')
        ```

        :param filename: filename, may contain relative or absolute path.
        :param populations: if `True`, population data will be saved.
        :param projections: if `True`, projection data will be saved.
        """
        IO.save(filename, populations, projections, self.id)

    ###################################
    # Memory
    ###################################
    def _cpp_memory_footprint(self):
        """
        Print the C++ memory consumption for populations, projections on the console.
        """
        for pop in self._data.populations:
            print(pop.name, pop._size_in_bytes())

        for proj in self._data.projections:
            print(proj.name, proj._size_in_bytes())

        for mon in self._data.monitors:
            print(type(mon), mon._size_in_bytes())

    ###################################
    # Access methods using names
    ###################################

    def get_population(self, name:str) -> "Population":
        """
        Returns the population with the given name.

        :param name: name of the population
        """
        for pop in self._data.populations:
            if pop.name == name:
                return pop
        Messages._print('get_population(): the population', name, 'does not exist in this network.')
        return None

    def get_projection(self, name:str) -> "Projection":
        """
        Returns the projection with the given name.

        :param name: name of the projection
        """
        for proj in self._data.projections:
            if proj.name == name:
                return proj
        Messages._print('get_projection(): the projection', name, 'does not exist in this network.')
        return None

    def get_monitor(self, name:str) -> "Monitor":
        """
        Returns the monitor with the given name.

        :param name: name of the monitor
        """
        for mon in self._data.monitors:
            if mon.name == name:
                return mon
        Messages._print('get_monitor(): the monitor', name, 'does not exist in this network.')
        return None

    def get_extension(self, name:str) :
        """
        Returns the extension with the given name.

        :param name: name of the extension
        """
        for ext in self._data.extensions:
            if ext.name == name:
                return ext
        Messages._print('get_extension(): the extension', name, 'does not exist in this network.')
        return None

    def get_constant(self, name:str) -> "Constant":
        """
        Returns the constant with the given name.

        :param name: name of the constant
        """
        for constant in self._data.constants:
            if constant.name == name:
                return constant
        Messages._print('get_constant(): the constant', name, 'does not exist in this network.')
        return None
    
    ###################################
    # Access methods for everything
    ###################################
    def get_populations(self) -> list["Population"]:
        """
        Returns a list of all declared populations in this network.
        """
        return self._data.populations

    def get_projections(self) -> list["Projection"]:
        """
        Returns a list of all declared projections for the current network. 
        """
        return self._data.projections

    def get_monitors(self) -> list["Monitor"]:
        """
        Returns a list of declared monitors. 
        """
        return self._data.monitors

    def get_extensions(self) -> list:
        """
        Returns a list of declared extensions (e.g. BOLD monitors). 
        """
        return self._data.extensions

    def get_constants(self) -> list:
        """
        Returns a list of declared constants. 
        """
        return self._data.constants


    
    ###################################
    # Store objects in the data structure
    ###################################
    def _add_population(self, population:Population) -> int :
        pop_id = len(self._data.populations)
        self._data.populations.append(population)
        return pop_id
    
    def _add_projection(self, projection:Projection) -> int :
        proj_id = len(self._data.projections)
        self._data.projections.append(projection)
        return proj_id
    
    def _add_monitor(self, monitor:Monitor) -> int :
        mon_id = len(self._data.monitors)
        self._data.monitors.append(monitor)
        return mon_id
    
    def _add_extension(self, extension) -> int :
        ext_id = len(self._data.extensions)
        self._data.extensions.append(extension)
        return ext_id
    
    def _add_constant(self, constant) -> int :
        const_id = len(self._data.constants)
        self._data.constants.append(constant)
        return const_id
    

    
    ###################################
    # Properties
    ###################################
    @property
    def instance(self):
        return self._data.instance
    
    @instance.setter
    def instance(self, instance) -> None:
        self._data.instance = instance

    @property
    def compiled(self) -> bool:
        """
        Boolean indicating whether the network has been compiled.
        """
        return self._data.compiled
    
    @compiled.setter
    def compiled(self, compilation_state: bool) -> None:
        self._data.compiled = compilation_state

    @property
    def directory(self) -> str:
        """
        Directory in which the network has been compiled.
        """
        return self._data.directory
    
    @directory.setter
    def directory(self, directory: str) -> None:
        self._data.directory = directory

    @property
    def time(self) -> float:
        "Current time t in milliseconds."
        return Global.get_time(self.id)

    @time.setter
    def time(self, t:float) -> None:
        Global.set_time(t, self.id)

    @property
    def current_step(self) -> int:
        "Current simulation step."
        return Global.get_current_step(self.id)

    @current_step.setter
    def current_step(self, t:int) -> None:
        Global.set_current_step(t, self.id)

    ################################
    ## @every decorator
    ################################

    def callbacks_enabled(self):
        """
        Returns True if callbacks are enabled for the network.
        """
        return self._callbacks_enabled

    def disable_callbacks(self):
        """
        Disables all callbacks for the network.
        """
        self._callbacks_enabled = False

    def enable_callbacks(self):
        """
        Enables all declared callbacks for the network.
        """
        self._callbacks_enabled = True

    def clear_all_callbacks(self):
        """
        Clears the list of declared callbacks for the network.

        Cannot be undone!
        """
        self._callbacks.clear()




from ANNarchy.intern import Messages

##################################
### Main method
##################################

def report(network:"Network", filename:str="./report.tex", standalone:bool=True, gather_subprojections:bool=False, title:str=None, author:str=None, date:str=None):
    """ 
    Generates a report describing the network.

    If the filename ends with ``.tex``, the TeX file is generated according to:

    > Nordlie E, Gewaltig M-O, Plesser HE (2009). Towards Reproducible Descriptions of Neuronal Network Models. PLoS Comput Biol 5(8): e1000456.

    If the filename ends with ``.md``, a (more complete) Markdown file is generated, which can be converted to pdf or html by ``pandoc``::

        pandoc report.md -o report.pdf
        pandoc report.md -o report.html

    :param network: Network instance.
    :param filename: name of the file where the report will be written.
    :param standalone: tells if the generated TeX file should be directly compilable or only includable (default: True). Ignored for Markdown.
    :param gather_subprojections: if a projection between two populations has been implemented as a multiple of projections between sub-populations, this flag allows to group them in the summary.
    :param title: title of the document (Markdown only).
    :param author: author of the document (Markdown only).
    :param date: date of the document (Markdown only).
    """

    if filename.endswith('.tex'):
        from .LatexReport import report_latex
        report_latex(filename, standalone, gather_subprojections, network.id)

    elif filename.endswith('.md'):
        from .MarkdownReport import report_markdown
        report_markdown(filename, standalone, gather_subprojections, title, author, date, network.id)

    else:
        Messages._error('report(): the filename must end with .tex or .md.')


"""
ANNarchy uses a distinct set of weights for each post-synaptic neuron (fully-connected).
However, operations like pooling or convolution are applied with the same weights for all pre-neurons within
a projection. Consequently, the required weight matrix need to be stored only once. This can
save computation time and especially memory space.

Contained classes:

* Convolution: simple-, layer-wise and bank of filter convolution
* Pooling: pooling operation (max, mean, etc)
* CopyProjection: mirrors the weights of another projection.

Implementation note:

* the implementation code was formerly in weightsharing.SharedProjection.
"""
from .Pooling import Pooling
from .Convolve import Convolution
from .Copy import Copy
from .Transpose import Transpose

# Export only the instantiable classes
__all__ = ['Pooling', 'Convolution', 'Copy', 'Transpose']



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.PopulationView import PopulationView

from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern import Messages

import re
import subprocess
import sys

def sort_odes(desc, locality='local'):
    equations = []
    is_ode = False
    for param in desc['variables']:
        if param['cpp'] == '':
            continue
        if param['method'] == 'event-driven':
            continue
        if param['name'] in desc[locality]:
            if param['switch']: # ODE
                if is_ode: # was already ODE
                    if len(equations) == 0:
                        equations.append(('ode', [param]))
                    else:
                        equations[-1][1].append(param)
                else: # new block
                    is_ode = True
                    equations.append(('ode', [param]))
            else: # non-ODE
                if is_ode:
                    is_ode = False
                    equations.append(('non-ode', [param]))
                else:
                    if len(equations) == 0:
                        equations.append(('non-ode', [param]))
                    else:
                        equations[-1][1].append(param)

    return equations

def generate_bound_code(param, obj):
    code = ""
    for bound, val in param['bounds'].items():
        if bound in ['min', 'max']:
            code += """if(%(var)s%(index)s %(operator)s %(val)s)
    %(var)s%(index)s = %(val)s;
""" % {
        'index': '%(local_index)s' if param['locality'] == 'local' else ('%(semiglobal_index)s' if param['locality'] == 'semiglobal' else'%(global_index)s'),
        'var' : param['name'], 'val' : val,
        'operator': '<' if bound=='min' else '>'
    }
    return code

def append_mask_multiplier(switch_code, var_name):
    """
    To apply SIMD vectorization on neural and continous synaptic updates the
    inner loop code should not contain if-else statements.

    Therefore, we add to switch statements of the ODE a multiplicant which
    represents the boolean condition.
    """
    return switch_code.replace(";", "* "+var_name+"%(local_index)s;")

def generate_non_ODE_block(variables, locality, obj, conductance_only, wrap_w, mask_variable, split_loop=False):
    " TODO: documentation "
    block_code = ""
    block_bounds = ""
    for param in variables:
        if conductance_only: # skip the variables which do not start with g_
            if not param['name'].startswith('g_'):
                continue

        # Add refractoriness
        if not param['name'].startswith('g_'):
            cpp_code = param['cpp'] if mask_variable is None else "if ("+mask_variable+"%(local_index)s) { "+param['cpp']+" }"
        else:
            cpp_code = param['cpp']

        bounds = generate_bound_code(param, obj)
        if wrap_w and param['name'] == "w":
            block_code += """
%(comment)s
if(%(wrap)s){
%(cpp)s
%(bounds)s
}
""" % { 'comment': '// ' + param['eq'],
        'cpp': cpp_code,
        'wrap': wrap_w,
        'bounds': bounds if not split_loop else ""}
            block_bounds += bounds if split_loop else ""

        else:
            block_code += """
%(comment)s
%(cpp)s
%(bounds)s
""" % { 'comment': '// ' + param['eq'],
        'cpp': cpp_code,
        'bounds': bounds if not split_loop else "" }
            block_bounds += bounds if split_loop else ""

    if not split_loop:
        return block_code
    else:
        return block_code, block_bounds


def generate_ODE_block(odes, locality, obj, conductance_only, wrap_w, mask_variable):
    code = ""

    # Count how many steps (midpoint has more than one step)
    nb_step = 0
    for param in odes:
        if isinstance(param['cpp'], list):
            nb_step = max(len(param['cpp']), nb_step)
        else:
            nb_step = max(1, nb_step)

    if len(odes) == 0:
        return ""

    # Iterate over all steps
    for step in range(nb_step):
        for param in odes:
            if conductance_only: # skip the variables which do not start with g_
                if not param['name'].startswith('g_'):
                    continue

            # Retrieve equation
            if isinstance(param['cpp'], list) and step < len(param['cpp']):
                eq = param['cpp'][step]
            elif isinstance(param['cpp'], str) and step == 0:
                eq = param['cpp']
            else:
                eq = ''
            # Generate code
            code += """
%(comment)s
%(cpp)s
""" % { 'comment': '// '+param['eq'],
        'cpp': eq }

    # Generate the switch code
    for param in odes:
        if conductance_only: # skip the variables which do not start with g_
            if not param['name'].startswith('g_'):
                continue

        bounds = generate_bound_code(param, obj)

        if not param['name'].startswith('g_'):
            switch = param['switch'] if mask_variable is None else append_mask_multiplier(param['switch'], mask_variable)
        else:
            switch = param['switch']

        if wrap_w and param['name'] == "w":
            code += """
%(comment)s
if(%(wrap)s){
%(switch)s
%(bounds)s
}
""" % { 'comment': '// '+param['eq'],
        'wrap': wrap_w,
        'bounds': bounds,
        'switch' : switch}
        else:
            code += """
%(comment)s
%(switch)s
%(bounds)s
""" % { 'comment': '// '+param['eq'],
        'bounds': bounds,
        'switch' : switch}


    return code

def generate_equation_code(desc, locality='local', obj='pop', conductance_only=False, wrap_w=None, mask_variable=None, padding=3):
    """
    Generate the C++ code for all equations part of *desc*.

    :param obj: either *obj* or *proj*
    :param locality: either *local*/*global* for populations or *local*/*semiglobal*/*global* for projections
    :param conductance_only: special case for populations considering only statements related to g_target variables
    :param wrap_w: if True, a condition check is added on the weight update
    :param with_refractory: special code generation for spiking models using refractoriness to ensure SIMD friendliness
    """
    # Separate ODEs from the pre- and post- equations
    odes = sort_odes(desc, locality)

    if odes == []: # No equations
        return ""

    # Generate code
    code = ""
    for type_block, block in odes:
        if type_block == 'ode':
            code += generate_ODE_block(block, locality, obj, conductance_only, wrap_w, mask_variable=mask_variable)
        elif type_block == 'non-ode':
            code += generate_non_ODE_block(block, locality, obj, conductance_only, wrap_w, mask_variable=mask_variable, split_loop=False)
        else:
            raise NotImplementedError

    # Add the padding to each line
    padded_code = tabify(code, padding)

    return padded_code

def determine_idx_type_for_projection(proj):
    """
    The suitable index type depends on the maximum number of neurons in
    pre-synaptic and post-synaptic layer.

    Notice (8th June 2021):

    It appears to a problem for the current Cython version to handle
    datatypes like "unsigned int". So I decided to replace the unsigned
    datatypes by an own definition. These definitions are placed in

    *ANNarchy/generator/Template/PyxTemplate.py*
    """
    # The user disabled this optimization.
    if ConfigManager().get('only_int_idx_type', proj.net_id):
        return "int", "int", "int", "int"

    # Currently only implemented for some cases,
    # the others default to "old" configuration
    if proj.synapse_type.type == "spike":
        return "int", "int", "int", "int"

    if _check_paradigm("cuda", proj.net_id):
        return "int", "int", "int", "int"

    if proj._storage_format != "lil" and ConfigManager().get('num_threads', proj.net_id)>1:
        return "int", "int", "int", "int"

    # max_size is related to the population sizes. As we use one type for
    # both dimension we need to determine the maximum
    pre_size = proj.pre.population.size if isinstance(proj.pre, PopulationView) else proj.pre.size
    post_size = proj.post.population.size if isinstance(proj.post, PopulationView) else proj.post.size
    max_size_one_dim = max(pre_size, post_size)
    max_size_both_dim = pre_size * post_size

    # For type decision we rely on the C++ boundaries which are decremented by 1
    # to allow usage of CSR-like formats without row overflow.
    if max_size_one_dim < 255:
        # 1 byte
        cpp_idx_type = "unsigned char"
        cython_idx_type= "_ann_uint8"

        if max_size_both_dim < 255:
            # can use the same type (should be seldom ...)
            cpp_size_type = "unsigned char"
            cython_size_type= "_ann_uint8"
        else:
            # next higher data type
            cpp_size_type = "unsigned short int"
            cython_size_type= "_ann_uint16"

    elif max_size_one_dim < 65534:
        # 2 byte
        cpp_idx_type = "unsigned short int"
        cython_idx_type= "_ann_uint16"

        if max_size_both_dim < 65534:
            cpp_size_type = "unsigned short int"
            cython_size_type= "_ann_uint16"
        else:
            cpp_size_type = "unsigned int"
            cython_size_type= "_ann_uint32"

    elif max_size_one_dim < 4294967294:
        # 4 byte
        cpp_idx_type = "unsigned int"
        cython_idx_type= "_ann_uint32"

        if max_size_both_dim < 4294967294:
            cpp_size_type = "unsigned int"
            cython_size_type= "_ann_uint32"
        else:
            cpp_size_type = "unsigned long int"
            cython_size_type= "_ann_uint64"

    else:
        # this is a hypothetical case I guess (HD: 4th June 2021)
        raise NotImplementedError("The matrix dimension exceeded the representable size ...")

    return cpp_idx_type, cython_idx_type, cpp_size_type, cython_size_type

def cpp_connector_available(connector_name, desired_format, storage_order, net_id):
    """
    Checks if a CPP implementation is available for the desired connection pattern
    (*connector_name*) and the target sparse matrix format (*desired_format*). Please
    note that not all formats are available for *pre_to_post* storage order.
    """
    # The user disabled this feature
    if not ConfigManager().get('use_cpp_connectors', net_id):
        return False

    cpp_patterns = {
        'st': {
            'post_to_pre': {
                "lil": ["Random", "Random Convergent", "All-to-All"],
                "csr": ["Random", "Random Convergent"],
                "coo": [],
                "hyb": [],
                "ell": [],
                "dense": ["Random"]
            },
            'pre_to_post': {
                "csr": ["Random", "Random Convergent"]
            }
        },
        'omp': {
            'post_to_pre': {
                "lil": ["Random"],
                "csr": [],
                "coo": [],
                "ell": []
            },
            'pre_to_post': {
                "csr": ["Random"]
            }
        },
        'cuda': {
            'post_to_pre': {
                "csr": ["Random", "Random Convergent"],
                "coo": [],
                "ellr": ["Random", "Random Convergent"],
                "dense": ["Random"]
            }
        }
    }

    if _check_paradigm("openmp", net_id):
        paradigm = "st" if ConfigManager().get('num_threads', net_id) == 1 else "omp"
    else:
        paradigm = "cuda"

    try:
        return connector_name in cpp_patterns[paradigm][storage_order][desired_format]

    except KeyError:
        # Fall back to Python construction
        return False

#####################################################################
#   Code formatting
#####################################################################
def indentLine(line, spaces=1):
    return (' ' * 4 * spaces) + line

def tabify(s, numSpaces):
    s = s.split('\n')
    s = map(lambda a, ns=numSpaces: indentLine(a, ns), s)
    s = '\n'.join(s)
    return s

def remove_trailing_spaces(code):
    """
    The generated code templates often contain empty lines, which are indented by tabify() or indentLine()
    afterwards which this introduces many white spaces which are annoying in some editors. The call of rstrip()
    on the complete string can not remove them. Therefore we implement this little helper function to call
    rstrip on each line.
    """
    stripped_lines = [line.rstrip() for line in code.split('\n')]

    stripped_code = ""
    for line in stripped_lines:
        stripped_code += line +'\n'

    return stripped_code

#####################################################################
#   Hardware-related
#####################################################################
def check_cuda_version(nvcc_executable):
    """
    Some features like atomic add for double values and power function are dependent on the CUDA version.
    """
    version_str = str(subprocess.check_output([nvcc_executable, "--version"]))
    try:
        version = float(version_str.split("\\")[-2].split(",")[1].split(" ")[2])
    except:
        try:
            version = float(version_str.split("\\")[-3].split(",")[1].split(" ")[2])
        except:
            Messages._error("Could not detect CUDA version: please check the CUDA installation or the configuration in annarchy.json")

    return version

def check_and_apply_pow_fix(eqs, cuda_version):
    """
    CUDA SDKs before 7.5 had an error if std=c++11 is enabled related
    to pow(double, int). Only pow(double, double) was detected as
    device function, the pow(double, int) will be detected as host
    function. (This was fixed within SDK 7.5)

    To support also earlier versions, we simply add a double type cast.
    """
    if eqs.strip() == "":
        # nothing to do
        return eqs

    if cuda_version > 7.0:
        # nothing to do, is working in higher SDKs
        return eqs

    # detect all pow statements
    pow_occur = re.findall(r"pow[\( [\S\s]*?\)*?, \d+\)]*?", eqs)
    for term in pow_occur:
        eqs = eqs.replace(term, term.replace(', ', ', (double)'))

    return eqs

def check_simd_instructions_type(simd_instr_set="avx", net_id:int=0):
    """
    Check the present CPUs if they offer a specific SIMD instruction set. We use 'lscpu' to detect
    the availabe instruction sets. In ANNarchy we support for now:

    - "avx":     256-bit register width
    - "avx512f": 512-bit register width

    Please note, that even though we detect the instruction set. Compiler flags must be set accordingly
    that it can work (i. e. either march=native detect the CPU correct or the -mavx/-mavx512f flags must be set)

    Parameters:

    * simd_instr_set: either "sse4_1", "avx" or "avx512f".

    Returns:

    True, if the specified flag was found, in any other cases it defaults to False.

    Remark (31th May 2021):

    This is a rather simple approach to detect the AVX capability of a CPU. If it fails, one can
    still hope for the auto-vectorization.
    """
    # Checks whether the user disabled the generation of SIMD code
    if ConfigManager().get('disable_SIMD_SpMV', net_id):
        return False

    # The hand-written codes are only validated
    # with g++ as target compiler
    if not sys.platform.startswith('linux'):
        return False

    try:
        # search for CPU flags
        lscpu_txt = (subprocess.check_output("lscpu | grep 'Flags' ", shell=True).strip()).decode()
        return simd_instr_set in lscpu_txt

    except:
        try:
            # lets try german
            lscpu_txt = (subprocess.check_output("lscpu | grep 'Markierungen' ", shell=True).strip()).decode()
            return simd_instr_set in lscpu_txt

        except:
            # give up and proceed without AVX
            return False

def get_highest_available_simd_instructions_type(net_id:int=0):
    """
    Check if SIMD operations are available. As higher order methods
    always contain the lower, we need to test in order SSE4, AVX, AVX512
    """
    simd_type = None

    if check_simd_instructions_type("sse4_1", net_id):
        simd_type = "sse"

    if check_simd_instructions_type("avx", net_id):
        simd_type = "avx"

    if check_simd_instructions_type("avx512f", net_id):
        simd_type = "avx512"

    return simd_type



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.core.Population import Population
from ANNarchy.core import Global

from ANNarchy.intern.Profiler import Profiler
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages

from math import ceil
import time
import tqdm
import operator

__all__ = [
    "simulate",
    "simulate_until",
    "step",
    "callbacks_enabled",
    "disable_callbacks",
    "enable_callbacks",
    "clear_all_callbacks",
    "every"
]

def simulate(
        duration:float, 
        measure_time:bool=False, 
        callbacks:bool=True, 
        net_id:int=0) -> None:
    """
    Simulates the network for the given duration in milliseconds. 
    
    The number of simulation steps is computed relative to the discretization step ``dt`` declared in ``setup()`` (default: 1ms):

    ```python
    ann.simulate(1000.0)
    ```

    :param duration: the duration in milliseconds.
    :param measure_time: defines whether the simulation time should be printed. 
    :param callbacks: defines if the callback methods (decorator ``every``) should be called.
    """
    if Profiler().enabled:
        t0 = time.time()
    
    # Access the network
    network = NetworkManager().get_network(net_id=net_id)

    # Sanity checks
    if not network.compiled:
        Messages._error('simulate(): the network is not compiled yet.')
    if not network.instance:
        Messages._error('simulate(): the network is not initialized yet.')

    # Compute the number of steps
    nb_steps = ceil(float(duration) / ConfigManager().get("dt", net_id))

    if measure_time:
        tstart = time.time()

    if callbacks and network._callbacks_enabled and len(network._callbacks) > 0:
        _simulate_with_callbacks(duration, net_id)
    else:
        batch = 1000 # someday find a better solution...
        if nb_steps < batch:
            network.instance.run(nb_steps)
        else:
            nb = int(nb_steps/batch)
            rest = nb_steps % batch

            for i in range(nb):
                network.instance.run(batch)

            if rest > 0:
                network.instance.run(rest)

    if measure_time:
        if net_id > 0:
            Messages._print('Simulating', duration/1000.0, 'seconds of the network', net_id, 'took', time.time() - tstart, 'seconds.')
        else:
            Messages._print('Simulating', duration/1000.0, 'seconds of the network took', time.time() - tstart, 'seconds.')

    # Store the Python and C++ timings. Please note, that the C++ core
    # measures in ms and Python measures in s
    if Profiler().enabled:
        t1 = time.time()
        Profiler().add_entry( t0, t1, "simulate", "simulate")

        # network single step
        overall_avg = Profiler()._cpp_profiler.get_avg_time("network", "step")
        Profiler().add_entry(overall_avg * nb_steps, 100.0, "overall", "cpp core")

        # single operations for populations
        for pop in network.get_populations():
            for func in ["step", "rng", "delay", "spike"]:
                avg_time = Profiler()._cpp_profiler.get_avg_time(pop.name, func)
                Profiler().add_entry( avg_time * nb_steps, (avg_time/overall_avg)*100.0, pop.name+"_"+func, "cpp core")

        # single operations for projections
        for proj in network.get_projections():
            for func in ["psp", "step", "post_event"]:
                avg_time = Profiler()._cpp_profiler.get_avg_time(proj.name, func)
                Profiler().add_entry( avg_time * nb_steps, (avg_time/overall_avg)*100.0, proj.name+"_"+func, "cpp core")

        monitor_avg = Profiler()._cpp_profiler.get_avg_time("network", "record")
        Profiler().add_entry( monitor_avg * nb_steps, (monitor_avg/overall_avg)*100.0, "record", "cpp core")

def simulate_until(max_duration:float, population: Population | list[Population], operator='and', measure_time:bool = False, net_id:int=0):
    """
    Runs the network for the maximal duration in milliseconds. If the ``stop_condition`` defined in the population becomes true during the simulation, it is stopped.

    One can specify several populations. If the stop condition is true for any of the populations, the simulation will stop ('or' function).

    Returns the actual duration of the simulation in milliseconds.

    Example:

    ```python
    pop1 = ann.Population( ..., stop_condition = "r > 1.0 : any")
    ann.compile()
    duration = ann.simulate_until(max_duration=1000.0, population=pop1)
    ```

    :param max_duration: Maximum duration of the simulation in milliseconds.
    :param population: (list of) population(s) whose ``stop_condition`` should be checked to stop the simulation.
    :param operator: Operator to be used ('and' or 'or') when multiple populations are provided (default: 'and').
    :param measure_time: Defines whether the simulation time should be printed (default=False).
    """
    # Access the network
    network = NetworkManager().get_network(net_id=net_id)

    # Sanity checks
    if not network.compiled:
        Messages._error('simulate_until(): the network is not compiled yet.')
    if not network.instance:
        Messages._error('simulate_until(): the network is not initialized yet.')

    # Compute maximum number of steps
    nb_steps = ceil(float(max_duration) / ConfigManager().get("dt", net_id))

    if not isinstance(population, list):
        population = [population]

    # Perform the simulation until max_duration is reached or the conditio is fulfilled.
    if measure_time:
        tstart = time.time()

    nb = network.instance.run_until(nb_steps, [pop.id for pop in population], True if operator=='and' else False)

    sim_time = float(nb) / ConfigManager().get("dt", net_id)
    if measure_time:
        Messages._print('Simulating', nb/ConfigManager().get("dt", net_id)/1000.0, 'seconds of the network took', time.time() - tstart, 'seconds.')
    return sim_time

def step(net_id=0):
    """
    Performs a single simulation step (duration = `dt`).
    """
    # Access the network
    network = NetworkManager().get_network(net_id=net_id)

    # Sanity check
    if not network.compiled:
        Messages._error('step(): the network is not compiled yet.')
    if not network.instance:
        Messages._error('step(): the network is not initialized yet.')

    # Simulate a single step
    network.instance.step()


################################
## Decorators
################################

def callbacks_enabled(net_id=0):
    """
    Returns True if callbacks are enabled for the network.
    """
    # Access the network
    network = NetworkManager().get_network(net_id=net_id)
    return network._callbacks_enabled

def disable_callbacks(net_id=0):
    """
    Disables all callbacks for the network.
    """
    # Access the network
    network = NetworkManager().get_network(net_id=net_id)
    network._callbacks_enabled = False

def enable_callbacks(net_id=0):
    """
    Enables all declared callbacks for the network.
    """
    # Access the network
    network = NetworkManager().get_network(net_id=net_id)
    network._callbacks_enabled = True

def clear_all_callbacks(net_id=0):
    """
    Clears the list of declared callbacks for the network.

    Cannot be undone!
    """
    # Access the network
    network = NetworkManager().get_network(net_id=net_id)
    network._callbacks.clear()


class every :
    """
    Decorator to declare a callback method that will be called periodically during the simulation.

    Example of setting increasing inputs to a population every 100 ms, with an offset of 90 ms (or -10 ms relative to the period):

    ```python
    net = ann.Network()

    @ann.every(network=net, period=100., offset=-10.)
    def step_input(n):
        pop.I = float(n) / 100.

    net.simulate(10000.)
    ```

    ``step_input()`` will be called at times 90, 190, ..., 9990 ms during the call to ``simulate()``.

    The method must accept only ``n`` as parameter (an integer being 0 the first time the method is called, and incremented afterwards) and can not return anything.

    The times at which the method is called are relative to the time when ``simulate()`` is called (if ``t`` is already 150 before calling ``simulate()``, the first call will then be made at ``t=240`` with the previous example).

    If multiple callbacks are defined, they will be called in the order of their declaration if they occur at the same time.

    ``wait`` can be combined with ``offset``, so if ``period=100.``, ``offset=50.`` and ``wait=500.``, the first call will be made 550 ms after the call to ``simulate()`
    
    :param network: the network instance that will catch the callbacks. By default it is the top-level network of id 0.
    :param period: interval in ms between two calls to the function. If less than ``dt``, will be called every step.
    :param offset: by default, the first call to the method will be made at the start of the simulation. The offset delays the call within the period (default: 0.0). Can be negative, in which case it will be counted from the end of the period.
    :param wait: allows to wait for a certain amount of time (in ms) before starting to call the method.

    """

    def __init__(self, network:"Network"=None, period:float=1.0, offset:float=0., wait:float=0.0) -> None:

        self.network = network if network is not None else NetworkManager().magic_network()
        self.network._callbacks.append(self)

        self.period = max(float(period), self.network.dt)
        self.offset = min(float(offset), self.period)
        self.wait = max(float(wait), 0.0)

    def __call__(self, f):
        
        # If there are decorator arguments, __call__() is only called
        # once, as part of the decoration process! You can only give
        # it a single argument, which is the function object.
        self.func = f
        return f


def _simulate_with_callbacks(duration, net_id=0):
    """
    Replaces simulate() when call_backs are defined.
    """

    # Access the network
    network = NetworkManager().get_network(net_id=net_id)

    # Duration
    t_start = Global.get_current_step(net_id)
    length = int(duration/network.dt)

    # Compute the times
    times = []
    for c in network._callbacks:
        period = int(c.period/network.dt)
        offset = int(c.offset/network.dt) % period
        wait = int(c.wait/network.dt)

        moments = range(t_start + wait + offset, t_start + length, period)
        n = 0
        for m in moments:
            times.append((m, c, n))
            n += 1

    # Sort the times to be sure they are in the right order.
    times = sorted(times, key=operator.itemgetter(0))

    for time, callback, n in times:
        # Advance the simulation to the desired time
        if time != Global.get_current_step(net_id):
            network.instance.run(time-Global.get_current_step(net_id))
        # Call the callback
        callback.func(n)

    # Go to the end of the duration
    if Global.get_current_step(net_id) < t_start + length:
        network.instance.run(t_start + length - Global.get_current_step(net_id))



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.parser.Equation import Equation
from ANNarchy.parser.StringManipulation import *
from ANNarchy.parser.ITE import *
from ANNarchy.parser.Extraction import *
from ANNarchy.parser.CoupledEquations import CoupledEquations
from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages

def analyse_neuron(neuron, net_id):
    """
    Parses the structure and generates code snippets for the neuron type.

    It returns a ``description`` dictionary with the following fields:

    * 'object': 'neuron' by default, to distinguish it from 'synapse'
    * 'type': either 'rate' or 'spiking'
    * 'raw_parameters': provided field
    * 'raw_equations': provided field
    * 'raw_functions': provided field
    * 'raw_reset': provided field
    * 'raw_spike': provided field
    * 'refractory': provided field
    * 'parameters': list of parameters defined for the neuron type
    * 'variables': list of variables defined for the neuron type
    * 'functions': list of functions defined for the neuron type
    * 'attributes': list of names of all parameters and variables
    * 'local': list of names of parameters and variables which are local to each neuron
    * 'global': list of names of parameters and variables which are global to the population
    * 'targets': list of targets used in the equations
    * 'random_distributions': list of random number generators used in the neuron equations
    * 'global_operations': list of global operations (min/max/mean...) used in the equations (unused)
    * 'spike': when defined, contains the equations of the spike conditions and reset.

    Each parameter is a dictionary with the following elements:

    * 'bounds': unused
    * 'ctype': 'type of the parameter: 'float', 'double', 'int' or 'bool'
    * 'eq': original equation in text format
    * 'flags': list of flags provided after the :
    * 'init': initial value
    * 'locality': 'local' or 'global'
    * 'name': name of the parameter

    Each variable is a dictionary with the following elements:

    * 'bounds': dictionary of bounds ('init', 'min', 'max') provided after the :
    * 'cpp': C++ code snippet updating the variable
    * 'ctype': type of the variable: 'float', 'double', 'int' or 'bool'
    * 'dependencies': list of variable and parameter names on which the equation depends
    * 'eq': original equation in text format
    * 'flags': list of flags provided after the :
    * 'init': initial value
    * 'locality': 'local' or 'global'
    * 'method': numericalmethod for ODEs
    * 'name': name of the variable
    * 'pre_loop': ODEs have a pre_loop term for precomputing dt/tau. dict with 'name' and 'value'. type must be inferred.
    * 'switch': ODEs have a switch term
    * 'transformed_eq': same as eq, except special terms (sums, rds) are replaced with a temporary name
    * 'untouched': dictionary of special terms, with their new name as keys and replacement values as values.

    The 'spike' element (when present) is a dictionary containing:

    * 'spike_cond': the C++ code snippet containing the spike condition ("v%(local_index)s > v_T")
    * 'spike_cond_dependencies': list of variables/parameters on which the spike condition depends
    * 'spike_reset': a list of reset statements, each of them composed of :
        * 'constraint': either '' or 'unless_refractory'
        * 'cpp': C++ code snippet
        * 'dependencies': list of variables on which the reset statement depends
        * 'eq': original equation in text format
        * 'name': name of the reset variable

    """

    # Store basic information
    description = {
        'object': 'neuron',
        'type': neuron.type,
        'raw_parameters': neuron.parameters,
        'raw_equations': neuron.equations,
        'raw_functions': neuron.functions,
    }

    # Spiking neurons additionally store the spike condition, the reset statements and a refractory period
    if neuron.type == 'spike':
        description['raw_reset'] = neuron.reset
        description['raw_spike'] = neuron.spike
        description['raw_axon_spike'] = neuron.axon_spike
        description['raw_axon_reset'] = neuron.axon_reset
        description['refractory'] = neuron.refractory

    # Extract parameters and variables names
    parameters = extract_parameters(neuron.parameters, neuron.extra_values, 'neuron', net_id=net_id)
    variables = extract_variables(neuron.equations, 'neuron', net_id=net_id)
    description['parameters'] = parameters
    description['variables'] = variables

    # Make sure r is defined for rate-coded networks
    from ANNarchy.extensions.bold.BoldModel import BoldModel
    if isinstance(neuron, BoldModel):
        found = False
        for var in description['parameters'] + description['variables']:
            if var['name'] == 'r':
                found = True
        if not found:
            description['variables'].append(
                {
                    'name': 'r', 
                    'locality': 'local', 
                    'bounds': {}, 
                    'ctype': ConfigManager().get('precision', net_id),
                    'init': 0.0, 
                    'flags': [], 
                    'eq': '', 
                    'cpp': ""
                }
            )
    elif neuron.type == 'rate':
        for var in description['parameters'] + description['variables']:
            if var['name'] == 'r':
                break
        else:
            Messages._error('Rate-coded neurons must define the variable "r".')

    else: # spiking neurons define r by default, it contains the average FR if enabled
        for var in description['parameters'] + description['variables']:
            if var['name'] == 'r':
                Messages._error('Spiking neurons use the variable "r" for the average FR, use another name.')

        description['variables'].append(
            {
                'name': 'r', 
                'locality': 'local', 
                'bounds': {}, 
                'ctype': ConfigManager().get('precision', net_id),
                'init': 0.0, 
                'flags': [], 
                'eq': '', 
                'cpp': ""
            }
        )

    # Extract functions
    functions = extract_functions(description=neuron.functions, local_global=False, net_id=net_id)
    description['functions'] = functions

    # Build lists of all attributes (param + var), which are local or global
    attributes, local_var, global_var, _ = get_attributes(parameters, variables, neuron=True)

    # Test if attributes are declared only once
    if len(attributes) != len(list(set(attributes))):
        Messages._error('Attributes must be declared only once.', attributes)

    # Store the attributes
    description['attributes'] = attributes
    description['local'] = local_var
    description['semiglobal'] = [] # only for projections
    description['global'] = global_var

    # Extract all targets
    targets = sorted(list(set(extract_targets(variables))))
    description['targets'] = targets
    if neuron.type == 'spike': # Add a default reset behaviour for conductances
        for target in targets:
            found = False
            for var in description['variables']:
                if var['name'] == 'g_' + target:
                    found = True
                    break
            if not found:
                description['variables'].append(
                    {
                        'name': 'g_'+target, 'locality': 'local', 'bounds': {}, 'ctype': ConfigManager().get('precision', net_id),
                        'init': 0.0, 'flags': [], 'eq': 'g_' + target+ ' = 0.0'
                    }
                )
                description['attributes'].append('g_'+target)
                description['local'].append('g_'+target)

    # Extract RandomDistribution objects
    random_distributions = extract_randomdist(description, net_id=net_id)
    description['random_distributions'] = random_distributions

    # Extract the spike condition if any
    if neuron.type == 'spike':
        description['spike'] = extract_spike_variable(description, net_id=net_id)
        description['axon_spike'] = extract_axon_spike_condition(description, net_id=net_id)

    # Global operation TODO
    description['global_operations'] = []

    # The ODEs may be interdependent (implicit, midpoint), so they need to be passed explicitely to CoupledEquations
    concurrent_odes = []

    # Translate the equations to C++
    for variable in description['variables']:
        # Get the equation
        eq = variable['transformed_eq']
        if eq.strip() == "":
            continue

        # For ODEs, assignment and so on, we would like to count the number of floating operations (FLOPs)
        num_flops = 0

        # Special variables (sums, global operations, rd) are placed in untouched, so that Sympy ignores them
        untouched = {}

        # Dependencies must be gathered
        dependencies = []

        # Replace sum(target) with _sum_exc__[i]
        for target in description['targets']:
            # sum() is valid for all targets
            eq = re.sub(r'(?P<pre>[^\w.])sum\(\)', r'\1sum(__all__)', eq)
            # Replace sum(target) with __sum_target__
            eq = re.sub(r'sum\(\s*'+target+r'\s*\)', r'__sum_'+target+'__', eq)
            untouched['__sum_'+target+'__'] = '_sum_' + target + '%(local_index)s'

        # Extract global operations
        eq, untouched_globs, global_ops = extract_globalops_neuron(variable['name'], eq, description, net_id=net_id)

        # Add the untouched variables to the global list
        for name, val in untouched_globs.items():
            if not name in untouched.keys():
                untouched[name] = val
        # Add found global operations to the list
        for tmp in global_ops:
            # Prevent doublons
            doubled = False
            for tmp2 in description['global_operations']:
                if tmp['function'] == tmp2['function'] and tmp['variable'] == tmp2['variable']:
                    doubled = True
                    break
            if not doubled:
                description['global_operations'] += [tmp]

        # Extract if-then-else statements
        eq, condition = extract_ite(variable['name'], eq, description)

        # Find the numerical method if any
        method = find_method(variable)

        # Process the bounds
        if 'min' in variable['bounds'].keys():
            if isinstance(variable['bounds']['min'], str):
                translator = Equation(
                    variable['name'],
                    variable['bounds']['min'],
                    description,
                    type = 'return',
                    untouched = untouched,
                    net_id=net_id
                )
                variable['bounds']['min'] = translator.parse().replace(';', '')
                dependencies += translator.dependencies()

        if 'max' in variable['bounds'].keys():
            if isinstance(variable['bounds']['max'], str):
                translator = Equation(
                    variable['name'],
                    variable['bounds']['max'],
                    description,
                    type = 'return',
                    untouched = untouched,
                    net_id=net_id
                )
                variable['bounds']['max'] = translator.parse().replace(';', '')
                dependencies += translator.dependencies()

        # Analyse the equation
        if condition == []:# No if-then-else
            translator = Equation(
                variable['name'],
                eq,
                description,
                method = method,
                untouched = untouched,
                    net_id=net_id
            )
            code = translator.parse()
            dependencies += translator.dependencies()
            num_flops = translator.num_flops
        
        else: # An if-then-else statement
            code, deps = translate_ITE(
                        variable['name'],
                        eq,
                        condition,
                        description,
                        untouched )
            dependencies += deps

        # ODEs have a switch statement:
        #   double _r = (1.0 - r)/tau;
        #   r[i] += dt* _r;
        # while direct assignments are one-liners:
        #   r[i] = 1.0
        if isinstance(code, str):
            pre_loop = {}
            cpp_eq = code
            switch = None
        else: # ODE
            pre_loop = code[0]
            cpp_eq = code[1]
            switch = code[2]

        # Replace untouched variables with their original name
        for prev, new in sorted(list(untouched.items()), key = lambda key : len(key[0]), reverse=True):
            if prev.startswith('g_'):
                cpp_eq = re.sub(r'([^_]+)'+prev, r'\1'+new, ' ' + cpp_eq).strip()
                if len(pre_loop) > 0:
                    pre_loop['value'] = re.sub(r'([^_]+)'+prev, new, ' ' + pre_loop['value']).strip()
                if switch:
                    switch = re.sub(r'([^_]+)'+prev, new, ' ' + switch).strip()
            else:
                cpp_eq = re.sub(prev, new, cpp_eq)
                if len(pre_loop) > 0:
                    pre_loop['value'] = re.sub(prev, new, pre_loop['value'])
                if switch:
                    switch = re.sub(prev, new, switch)

        # Replace local functions
        for f in description['functions']:
            cpp_eq = re.sub(r'([^\w]*)'+f['name']+r'\(',
                            r'\1'+f['name'] + '(', ' ' + cpp_eq).strip()

        # Store the result
        variable['pre_loop'] = pre_loop # Things to be declared before the for loop (eg. dt)
        variable['cpp'] = cpp_eq # the C++ equation
        variable['switch'] = switch # switch value of ODE
        variable['transformed_eq'] = eq # the equation with untouched terms
        variable['untouched'] = untouched # untouched vocabulary
        variable['method'] = method # numerical method
        variable['dependencies'] = list(set(dependencies)) # list of dependencies
        variable['num_flops'] = num_flops

        # If the method is implicit or midpoint, the equations must be solved concurrently (depend on v[t+1])
        if method in ['implicit', 'midpoint', 'rk4'] and switch is not None:
            concurrent_odes.append(variable)

    # After all variables are processed, do it again if they are concurrent
    if len(concurrent_odes) > 1 :
        solver = CoupledEquations(description, concurrent_odes, net_id)
        new_eqs = solver.parse()
        for idx, variable in enumerate(description['variables']):
            for new_eq in new_eqs:
                if variable['name'] == new_eq['name']:
                    description['variables'][idx] = new_eq

    return description
